#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pyqtgraph as pg
from pyqtgraph import PlotWidget
from PySide6.QtCore import QPoint, Signal, Slot
from PySide6.QtGui import QAction, QDoubleValidator, Qt
from PySide6.QtWidgets import QDialog, QMenu

from pyminflux.analysis import (
    assign_data_to_clusters,
    prepare_histogram,
    reassign_fluo_ids_by_majority_vote,
)
from pyminflux.processor import MinFluxProcessor
from pyminflux.state import State
from pyminflux.ui.helpers import export_plot_interactive
from pyminflux.ui.ui_color_unmixer import Ui_ColorUnmixer


class ColorUnmixer(QDialog, Ui_ColorUnmixer):
    """
    A QDialog to display the dcr histogram and to assigning the fluorophore ids.
    """

    # Signal that the fluorophore IDs have been assigned
    fluorophore_ids_assigned = Signal(int, name="fluorophore_ids_assigned")

    def __init__(self, processor: MinFluxProcessor, parent):
        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_ColorUnmixer()
        self.ui.setupUi(self)

        # Initialize state
        self.state = State()

        # Set up ui elements
        self.ui.pbAssign.setEnabled(False)
        self.ui.leBinSize.setValidator(QDoubleValidator(bottom=0.0))
        self.ui.leBinSize.setText(str(self.state.dcr_bin_size))
        self.ui.pbManualAssign.setEnabled(False)
        self.ui.leManualThreshold.setValidator(QDoubleValidator(bottom=0.0))
        self.ui.leManualThreshold.setText(str(self.state.dcr_manual_threshold))

        # Constants
        self.brush = pg.mkBrush(0, 0, 0, 255)
        self.pen = pg.mkPen(None)

        # Keep a reference to the Processor
        self.processor = processor

        # Keep track of the global data range and step
        self.n_dcr_max = None
        self.dcr_bin_edges = None
        self.dcr_bin_centers = None
        self.dcr_bin_width = None

        # Keep track of the fluorophore ID assignments
        self.assigned_fluorophore_ids = None

        # Create the plot elements
        self.plot_widget = PlotWidget(parent=self, background="w", title="dcr")

        # Plot the dcr histogram
        self.plot_dcr_histogram()

        # Add them to the UI
        self.ui.main_layout.addWidget(self.plot_widget)

        # Add connections
        self.ui.cbNumFluorophores.currentIndexChanged.connect(
            self.persist_num_fluorophores
        )
        self.ui.leBinSize.textChanged.connect(self.persist_dcr_bin_size)
        self.ui.leBinSize.editingFinished.connect(self.plot_dcr_histogram)
        self.ui.leManualThreshold.textChanged.connect(self.persist_dcr_manual_threshold)
        self.ui.pbDetect.clicked.connect(self.detect_fluorophores)
        self.ui.pbAssign.clicked.connect(self.assign_fluorophores_ids)
        self.ui.pbPreview.clicked.connect(self.preview_manual_assignment)
        self.ui.pbManualAssign.clicked.connect(self.assign_fluorophores_ids)

    @Slot(str, name="persist_dcr_bin_size")
    def persist_dcr_bin_size(self, text):
        try:
            dcr_bin_size = float(text)
        except ValueError as _:
            return
        self.state.dcr_bin_size = dcr_bin_size

    @Slot(str, name="persist_num_fluorophores")
    def persist_num_fluorophores(self, text):
        num_fluorophores = self.ui.cbNumFluorophores.currentIndex() + 1
        self.state.num_fluorophores = num_fluorophores
        self.ui.pbAssign.setEnabled(False)

    @Slot(str, name="persist_dcr_manual_threshold")
    def persist_dcr_manual_threshold(self, text):
        try:
            dcr_manual_threshold = float(text)
        except ValueError as _:
            return
        self.state.dcr_manual_threshold = float(dcr_manual_threshold)

    @Slot(None, name="plot_dcr_histogram")
    def plot_dcr_histogram(self):
        """Plot the dcr histogram. This is always performed assuming all data belongs to one fluorophore."""

        # Do we have something to plot?
        if self.processor is None or self.processor.full_dataframe is None:
            return

        if self.state.dcr_bin_size == 0:
            # Calculate the dcr histogram
            n_dcr, dcr_bin_edges, dcr_bin_centers, dcr_bin_width = prepare_histogram(
                self.processor.full_dataframe["dcr"].values,
                auto_bins=True,
            )
        else:
            # Calculate the dcr histogram
            n_dcr, dcr_bin_edges, dcr_bin_centers, dcr_bin_width = prepare_histogram(
                self.processor.full_dataframe["dcr"].values,
                auto_bins=False,
                bin_size=self.state.dcr_bin_size,
            )

        # Remember the global histogram range and step
        self.n_dcr_max = n_dcr.max()
        self.dcr_bin_edges = dcr_bin_edges
        self.dcr_bin_centers = dcr_bin_centers
        self.dcr_bin_width = dcr_bin_width

        # Update the state and the ui element
        self.state.dcr_bin_size = self.dcr_bin_width
        self.ui.leBinSize.setText(f"{self.dcr_bin_width:.5f}")

        # It one or more charts already exists, remove them
        for item in self.plot_widget.allChildItems():
            self.plot_widget.removeItem(item)

        # Create the bar chart
        chart = pg.BarGraphItem(
            x=self.dcr_bin_centers,
            height=n_dcr,
            width=0.9 * self.dcr_bin_width,
            brush=self.brush,
        )
        self.plot_widget.setMouseEnabled(x=False, y=False)
        self.plot_widget.setXRange(self.dcr_bin_centers[0], self.dcr_bin_centers[-1])
        self.plot_widget.setYRange(0.0, self.n_dcr_max)
        self.plot_widget.setLabel("bottom", "Complete dataset")
        self.plot_widget.setMenuEnabled(False)
        self.plot_widget.addItem(chart)
        self.plot_widget.scene().sigMouseClicked.connect(
            self.histogram_raise_context_menu
        )

    @Slot(None, name="preview_manual_assignment")
    def preview_manual_assignment(self):
        """Preview the manual assignment."""

        # Get the data
        dcr = self.processor.full_dataframe["dcr"].values
        if len(dcr) == 0:
            return

        # Get the threshold
        threshold = float(self.ui.leManualThreshold.text())

        # Keep track of the total number of elements for normalisation
        n_total = len(dcr)

        # Partition the data
        y_pred = dcr > threshold

        # Turn the prediction into a fluo_id
        fluo_ids = y_pred + 1

        # Reassign fluorophore IDs majority vote if necessary
        if self.state.num_fluorophores > 1:
            fluo_ids = reassign_fluo_ids_by_majority_vote(
                fluo_ids, self.processor.full_dataframe["tid"]
            )

        # It one or more charts already exists, remove them
        for item in self.plot_widget.allChildItems():
            self.plot_widget.removeItem(item)

        # Prepare some colors
        brushes = ["b", "r", "g", "m", "c"]

        # Calculate the bar width as a function of the number of fluorophores
        bar_width = 0.9 / self.state.num_fluorophores
        offset = self.dcr_bin_width / self.state.num_fluorophores

        # Create new histograms (always 2)
        for f in range(1, 3):
            data = dcr[fluo_ids == f]
            if len(data) == 0:
                continue

            # Calculate the dcr histogram using the global bins
            n_dcr, _ = np.histogram(data, bins=self.dcr_bin_edges, density=False)

            # Normalize by the total count
            n_dcr = n_dcr / n_total

            # Create the bar chart
            chart = pg.BarGraphItem(
                x=self.dcr_bin_centers + f * offset,
                height=n_dcr,
                width=bar_width * self.dcr_bin_width,
                brush=brushes[f - 1],
                alpha=0.5,
            )
            self.plot_widget.addItem(chart)

        # Store the predictions (1-shifted)
        self.assigned_fluorophore_ids = fluo_ids

        # Make sure to enable the assign button
        self.ui.pbManualAssign.setEnabled(True)

    @Slot(None, name="detect_fluorophores")
    def detect_fluorophores(self):
        """Detect fluorophores."""

        # Get the data
        dcr = self.processor.full_dataframe["dcr"].values
        if len(dcr) == 0:
            return

        # Fit the data to the requested number of clusters
        fluo_ids = assign_data_to_clusters(dcr, self.state.num_fluorophores, seed=42)

        # Reassign fluorophore IDs majority vote if necessary
        if self.state.num_fluorophores > 1:
            fluo_ids = reassign_fluo_ids_by_majority_vote(
                fluo_ids, self.processor.full_dataframe["tid"]
            )

        # It one or more charts already exists, remove them
        for item in self.plot_widget.allChildItems():
            self.plot_widget.removeItem(item)

        # Prepare some colors
        brushes = ["b", "r", "g", "m", "c"]

        # Calculate the bar width as a function of the number of fluorophores
        bar_width = 0.9 / self.state.num_fluorophores
        offset = self.dcr_bin_width / self.state.num_fluorophores

        # Keep track of the total number of values for histogrm normalization
        n_values = len(dcr)

        # Create new histograms
        for f in range(1, self.state.num_fluorophores + 1):
            data = dcr[fluo_ids == f]
            if len(data) == 0:
                continue

            # Calculate the dcr histogram using the global bins
            n_dcr, _ = np.histogram(data, bins=self.dcr_bin_edges, density=False)

            # Normalize by the total count
            n_dcr = n_dcr / n_values

            # Create the bar chart
            chart = pg.BarGraphItem(
                x=self.dcr_bin_centers + f * offset,
                height=n_dcr,
                width=bar_width * self.dcr_bin_width,
                brush=brushes[f - 1],
                alpha=0.5,
            )
            self.plot_widget.addItem(chart)

        # Store the predictions (1-shifted)
        self.assigned_fluorophore_ids = fluo_ids

        # Make sure to enable the assign button
        self.ui.pbAssign.setEnabled(True)

    @Slot(None, name="assign_fluorophores_ids")
    def assign_fluorophores_ids(self):
        """Assign the fluorophores ids."""

        if self.assigned_fluorophore_ids is None:
            return

        # Assign the IDs via the processor
        self.processor.set_fluorophore_ids(self.assigned_fluorophore_ids)

        # Inform that the fluorophore IDs have been assigned
        self.fluorophore_ids_assigned.emit(
            len(np.unique(self.assigned_fluorophore_ids))
        )

        # Disable the button until the new detection is run
        self.ui.pbAssign.setEnabled(False)

    def histogram_raise_context_menu(self, ev):
        """Create a context menu on the plot."""
        if ev.button() == Qt.MouseButton.RightButton:
            menu = QMenu()
            export_action = QAction("Export plot")
            export_action.triggered.connect(
                lambda checked: export_plot_interactive(ev.currentItem)
            )
            menu.addAction(export_action)
            pos = ev.screenPos()
            menu.exec(QPoint(int(pos.x()), int(pos.y())))
            ev.accept()
        else:
            ev.ignore()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from enum import IntEnum
from typing import Union

from ..base import Singleton
from ..reader import Metadata


class ColorCode(IntEnum):
    NONE = 0
    BY_TID = 1
    BY_FLUO = 2


class State(metaclass=Singleton):
    """State machine (singleton class).

    This keeps the State for the user interface! That is, only classes under pyminflux.ui
    should use State()! The core API must remain independent of State!
    """

    __SLOTS__ = [
        "plot_average_localisations",
        "color_code",
        "min_trace_length",
        "efo_expected_frequency",
        "efo_thresholds",
        "applied_efo_thresholds",
        "cfr_thresholds",
        "applied_cfr_thresholds",
        "efo_range",
        "cfr_range",
        "loc_precision_range",
        "enable_cfr_lower_threshold",
        "enable_cfr_upper_threshold",
        "cfr_threshold_factor",
        "efo_bin_size_hz",
        "weigh_avg_localization_by_eco",
        "x_param",
        "y_param",
        "num_fluorophores",
        "dcr_bin_size",
        "dcr_manual_threshold",
        "z_scaling_factor",
        "plot_export_dpi",
        "frc_lateral_resolution",
        "frc_temporal_resolution",
        "frc_num_repeats",
        "frc_endpoint_only",
        "open_console_at_start",
    ]

    def __init__(self):
        """Constructor.

        A singleton object of this dataclass acts as a state machine that allows various parts of the
        application to synchronize state among them.
        """

        # Plotting options
        self.plot_average_localisations: bool = False
        self.color_code: ColorCode = ColorCode.NONE

        # Minimum number of localizations to consider a trace
        self.min_trace_length: int = 1

        # EFO bin size in Hz (if 0.0, the bin size will be automatically estimated)
        self.efo_bin_size_hz: float = 1000.0

        # EFO expected frequency for single emitters
        self.efo_expected_frequency: float = 0.0

        # Lower and upper (absolute) thresholds for the EFO and CFR values
        self.efo_thresholds: Union[None, tuple] = None
        self.cfr_thresholds: Union[None, tuple] = None

        # Applied lower and upper (absolute) thresholds for the EFO and CFR values
        self.applied_efo_thresholds: Union[None, tuple] = None
        self.applied_cfr_thresholds: Union[None, tuple] = None

        # Histogram ranges
        self.efo_range: Union[None, tuple] = None
        self.cfr_range: Union[None, tuple] = None
        self.loc_precision_range: Union[None, tuple] = None

        # Parameter bounds
        self.enable_cfr_lower_threshold: bool = False
        self.enable_cfr_upper_threshold: bool = True

        # CFR thresholding parameters
        self.cfr_threshold_factor: float = 2.0

        # Weigh average localization by ECO
        self.weigh_avg_localization_by_eco: bool = False

        # Current parameter selection
        self.x_param: str = "x"
        self.y_param: str = "y"

        # Number of fluorophores in the sample
        self.num_fluorophores: int = 1
        self.dcr_bin_size: float = 0.0
        self.dcr_manual_threshold: float = 0.0

        # Z scaling factor
        self.z_scaling_factor: float = 0.7

        # Resolution for exporting plots as images
        self.plot_export_dpi: int = 600

        # FRC analysis
        self.frc_lateral_resolution: float = 4.0
        self.frc_temporal_resolution: float = 1800.0
        self.frc_num_repeats: int = 5
        self.frc_endpoint_only: bool = False

        # Whether to open the console at application start
        self.open_console_at_start: bool = False

    def asdict(self) -> dict:
        """Return class as dictionary."""
        return {
            "min_trace_length": self.min_trace_length,
            "plot_average_localisations": self.plot_average_localisations,
            "color_code": str(ColorCode(self.color_code)),
            "efo_expected_frequency": self.efo_expected_frequency,
            "efo_thresholds": self.efo_thresholds,
            "applied_efo_thresholds": self.applied_efo_thresholds,
            "cfr_thresholds": self.cfr_thresholds,
            "applied_cfr_threshold": self.applied_cfr_thresholds,
            "efo_range": self.efo_range,
            "cfr_range": self.cfr_range,
            "loc_precision_range": self.loc_precision_range,
            "enable_cfr_lower_threshold": self.enable_cfr_lower_threshold,
            "enable_cfr_upper_threshold": self.enable_cfr_upper_threshold,
            "cfr_threshold_factor": self.cfr_threshold_factor,
            "efo_bin_size_hz": self.efo_bin_size_hz,
            "weigh_avg_localization_by_eco": self.weigh_avg_localization_by_eco,
            "x_param": self.x_param,
            "y_param": self.y_param,
            "num_fluorophores": self.num_fluorophores,
            "dcr_bin_size": self.dcr_bin_size,
            "dcr_manual_threshold": self.dcr_manual_threshold,
            "z_scaling_factor": self.z_scaling_factor,
            "plot_export_dpi": self.plot_export_dpi,
            "frc_lateral_resolution": self.frc_lateral_resolution,
            "frc_temporal_resolution": self.frc_temporal_resolution,
            "frc_num_repeats": self.frc_num_repeats,
            "frc_endpoint_only": self.frc_endpoint_only,
            "open_console_at_start": self.open_console_at_start,
        }

    def reset(self):
        """Reset to data-specific settings."""

        self.efo_thresholds = None
        self.applied_efo_thresholds = None
        self.cfr_thresholds = None
        self.applied_cfr_thresholds = None

    def full_reset(self):
        """Reset to defaults."""

        self.min_trace_length = 1
        self.plot_average_localisations = False
        self.color_code: ColorCode = ColorCode.NONE
        self.efo_expected_frequency = 0.0
        self.efo_thresholds = None
        self.applied_efo_thresholds = None
        self.cfr_thresholds = None
        self.applied_cfr_thresholds = None
        self.efo_range = None
        self.cfr_range = None
        self.loc_precision_range = None
        self.enable_cfr_lower_threshold = False
        self.enable_cfr_upper_threshold = True
        self.cfr_threshold_factor = 2.0
        self.efo_bin_size_hz = 1000.0
        self.weigh_avg_localization_by_eco = False
        self.x_param = "x"
        self.y_param = "y"
        self.num_fluorophores = 1
        self.dcr_bin_size = 0.0
        self.dcr_manual_threshold = 0.0
        self.z_scaling_factor = 0.7
        self.plot_export_dpi = 600
        self.frc_lateral_resolution = 4.0
        self.frc_temporal_resolution = 1800.0
        self.frc_num_repeats = 5
        self.frc_endpoint_only = False
        self.open_console_at_start = False

    def update_from_metadata(self, metadata: Metadata):
        """Update State from the Metadata parameters from a `.pmx` file."""
        self.min_trace_length = metadata.min_trace_length
        self.efo_thresholds = metadata.efo_thresholds
        self.applied_efo_thresholds = metadata.efo_thresholds
        self.cfr_thresholds = metadata.cfr_thresholds
        self.applied_cfr_thresholds = metadata.cfr_thresholds
        self.num_fluorophores = metadata.num_fluorophores
        self.z_scaling_factor = metadata.z_scaling_factor

    def __str__(self):
        """Human-readable representation."""
        self.__repr__()

    def __repr__(self):
        """State representation."""
        return "\n".join([f"{key}: {value}" for key, value in self.asdict().items()])



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pyqtgraph as pg
from pyqtgraph import PlotWidget
from PySide6.QtCore import QPoint, Signal, Slot
from PySide6.QtGui import QAction, QColor, QFont, Qt
from PySide6.QtWidgets import QDialog, QMenu

from pyminflux.processor import MinFluxProcessor
from pyminflux.state import State
from pyminflux.ui.helpers import export_plot_interactive
from pyminflux.ui.trace_dataviewer import TraceDataViewer
from pyminflux.ui.ui_trace_stats_viewer import Ui_TraceStatsViewer


class TraceStatsViewer(QDialog, Ui_TraceStatsViewer):
    """
    A QDialog to display trace statistics.
    """

    export_trace_stats_requested = Signal(None, name="export_trace_stats_requested")

    def __init__(self, processor: MinFluxProcessor, parent):
        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_TraceStatsViewer()
        self.ui.setupUi(self)

        # Initialize state
        self.state = State()

        # Keep a reference to the Processor
        self.processor = processor

        # Create the trace data viewer (data will be automatically set)
        self.trace_dataviewer = TraceDataViewer()

        # Add it to the UI
        self.ui.main_layout.addWidget(self.trace_dataviewer)

        # Add connections
        self.ui.pbExport.clicked.connect(
            lambda _: self.export_trace_stats_requested.emit()
        )

        # Update the viewer
        self.update()

        # Show the data
        self.trace_dataviewer.show()

    @Slot(None, name="update")
    def update(self):
        """Update the viewer."""
        if self.processor is None:
            return
        self.trace_dataviewer.set_data(self.processor.filtered_dataframe_stats)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QAction,
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QGridLayout,
    QMainWindow,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QSplitter,
    QStatusBar,
    QWidget,
)


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1000, 800)
        MainWindow.setMinimumSize(QSize(640, 480))
        self.actionQuit = QAction(MainWindow)
        self.actionQuit.setObjectName("actionQuit")
        self.actionConsole = QAction(MainWindow)
        self.actionConsole.setObjectName("actionConsole")
        self.actionConsole.setCheckable(True)
        self.actionConsole.setChecked(False)
        self.actionData_viewer = QAction(MainWindow)
        self.actionData_viewer.setObjectName("actionData_viewer")
        self.actionData_viewer.setCheckable(True)
        self.actionData_viewer.setChecked(True)
        self.actionState = QAction(MainWindow)
        self.actionState.setObjectName("actionState")
        self.actionOptions = QAction(MainWindow)
        self.actionOptions.setObjectName("actionOptions")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionExport_data = QAction(MainWindow)
        self.actionExport_data.setObjectName("actionExport_data")
        self.actionManual = QAction(MainWindow)
        self.actionManual.setObjectName("actionManual")
        self.actionWebsite = QAction(MainWindow)
        self.actionWebsite.setObjectName("actionWebsite")
        self.actionCode_repository = QAction(MainWindow)
        self.actionCode_repository.setObjectName("actionCode_repository")
        self.actionIssues = QAction(MainWindow)
        self.actionIssues.setObjectName("actionIssues")
        self.actionMailing_list = QAction(MainWindow)
        self.actionMailing_list.setObjectName("actionMailing_list")
        self.actionExport_stats = QAction(MainWindow)
        self.actionExport_stats.setObjectName("actionExport_stats")
        self.actionCheck_for_updates = QAction(MainWindow)
        self.actionCheck_for_updates.setObjectName("actionCheck_for_updates")
        self.actionFRC_analyzer = QAction(MainWindow)
        self.actionFRC_analyzer.setObjectName("actionFRC_analyzer")
        self.actionWhat_s_new = QAction(MainWindow)
        self.actionWhat_s_new.setObjectName("actionWhat_s_new")
        self.actionUnmixer = QAction(MainWindow)
        self.actionUnmixer.setObjectName("actionUnmixer")
        self.actionTime_Inspector = QAction(MainWindow)
        self.actionTime_Inspector.setObjectName("actionTime_Inspector")
        self.actionAnalyzer = QAction(MainWindow)
        self.actionAnalyzer.setObjectName("actionAnalyzer")
        self.actionTrace_Stats_Viewer = QAction(MainWindow)
        self.actionTrace_Stats_Viewer.setObjectName("actionTrace_Stats_Viewer")
        self.actionTrace_Length_Viewer = QAction(MainWindow)
        self.actionTrace_Length_Viewer.setObjectName("actionTrace_Length_Viewer")
        self.actionLoad = QAction(MainWindow)
        self.actionLoad.setObjectName("actionLoad")
        self.actionSave = QAction(MainWindow)
        self.actionSave.setObjectName("actionSave")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.splitter_layout = QSplitter(self.centralwidget)
        self.splitter_layout.setObjectName("splitter_layout")
        self.splitter_layout.setMaximumSize(QSize(16777215, 16777215))
        self.splitter_layout.setOrientation(Qt.Vertical)

        self.gridLayout.addWidget(self.splitter_layout, 0, 0, 1, 1)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName("menubar")
        self.menubar.setGeometry(QRect(0, 0, 1000, 23))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuView = QMenu(self.menubar)
        self.menuView.setObjectName("menuView")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuAnalysis = QMenu(self.menubar)
        self.menuAnalysis.setObjectName("menuAnalysis")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuAnalysis.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionLoad)
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExport_data)
        self.menuFile.addAction(self.actionExport_stats)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionOptions)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menuView.addAction(self.actionData_viewer)
        self.menuView.addAction(self.actionConsole)
        self.menuView.addSeparator()
        self.menuView.addAction(self.actionState)
        self.menuHelp.addAction(self.actionManual)
        self.menuHelp.addAction(self.actionWebsite)
        self.menuHelp.addAction(self.actionCode_repository)
        self.menuHelp.addAction(self.actionIssues)
        self.menuHelp.addAction(self.actionMailing_list)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionWhat_s_new)
        self.menuHelp.addAction(self.actionCheck_for_updates)
        self.menuHelp.addAction(self.actionAbout)
        self.menuAnalysis.addAction(self.actionUnmixer)
        self.menuAnalysis.addAction(self.actionTime_Inspector)
        self.menuAnalysis.addAction(self.actionAnalyzer)
        self.menuAnalysis.addSeparator()
        self.menuAnalysis.addAction(self.actionTrace_Stats_Viewer)
        self.menuAnalysis.addAction(self.actionTrace_Length_Viewer)
        self.menuAnalysis.addAction(self.actionFRC_analyzer)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)

    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(
            QCoreApplication.translate("MainWindow", "MainWindow", None)
        )
        self.actionQuit.setText(QCoreApplication.translate("MainWindow", "Quit", None))
        # if QT_CONFIG(tooltip)
        self.actionQuit.setToolTip(
            QCoreApplication.translate("MainWindow", "Quit application", None)
        )
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionQuit.setShortcut("")
        # endif // QT_CONFIG(shortcut)
        self.actionConsole.setText(
            QCoreApplication.translate("MainWindow", "Console", None)
        )
        self.actionData_viewer.setText(
            QCoreApplication.translate("MainWindow", "Data viewer", None)
        )
        self.actionState.setText(
            QCoreApplication.translate("MainWindow", "[DEBUG] Show state", None)
        )
        self.actionOptions.setText(
            QCoreApplication.translate("MainWindow", "Options", None)
        )
        self.actionAbout.setText(
            QCoreApplication.translate("MainWindow", "About", None)
        )
        self.actionExport_data.setText(
            QCoreApplication.translate("MainWindow", "Export data", None)
        )
        self.actionManual.setText(
            QCoreApplication.translate("MainWindow", "Manual", None)
        )
        self.actionWebsite.setText(
            QCoreApplication.translate("MainWindow", "Website", None)
        )
        self.actionCode_repository.setText(
            QCoreApplication.translate("MainWindow", "Repository", None)
        )
        self.actionIssues.setText(
            QCoreApplication.translate("MainWindow", "Issues", None)
        )
        self.actionMailing_list.setText(
            QCoreApplication.translate("MainWindow", "Mailing list", None)
        )
        self.actionExport_stats.setText(
            QCoreApplication.translate("MainWindow", "Export trace stats", None)
        )
        self.actionCheck_for_updates.setText(
            QCoreApplication.translate("MainWindow", "Check for updates", None)
        )
        self.actionFRC_analyzer.setText(
            QCoreApplication.translate("MainWindow", "FRC Analyzer", None)
        )
        self.actionWhat_s_new.setText(
            QCoreApplication.translate("MainWindow", "What's new?", None)
        )
        self.actionUnmixer.setText(
            QCoreApplication.translate("MainWindow", "Unmixer", None)
        )
        self.actionTime_Inspector.setText(
            QCoreApplication.translate("MainWindow", "Time Inspector", None)
        )
        self.actionAnalyzer.setText(
            QCoreApplication.translate("MainWindow", "Analyzer", None)
        )
        self.actionTrace_Stats_Viewer.setText(
            QCoreApplication.translate("MainWindow", "Trace Stats Viewer", None)
        )
        self.actionTrace_Length_Viewer.setText(
            QCoreApplication.translate("MainWindow", "Trace Length Viewer", None)
        )
        self.actionLoad.setText(QCoreApplication.translate("MainWindow", "Load", None))
        # if QT_CONFIG(shortcut)
        self.actionLoad.setShortcut(
            QCoreApplication.translate("MainWindow", "Ctrl+L", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.actionSave.setText(QCoreApplication.translate("MainWindow", "Save", None))
        # if QT_CONFIG(shortcut)
        self.actionSave.setShortcut(
            QCoreApplication.translate("MainWindow", "Ctrl+S", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", "File", None))
        self.menuView.setTitle(QCoreApplication.translate("MainWindow", "View", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", "Help", None))
        self.menuAnalysis.setTitle(
            QCoreApplication.translate("MainWindow", "Analysis", None)
        )

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from pathlib import Path
from typing import Union

import h5py
import numpy as np

from pyminflux.processor import MinFluxProcessor
from pyminflux.state import State


class PyMinFluxNativeWriter:
    __docs__ = "Writer of (processed) MINFLUX into native `.pmx` format."

    def __init__(self, processor: MinFluxProcessor):
        self.processor = processor
        self._state = State()
        self._message = ""

    def write(self, file_name: Union[Path, str]) -> bool:

        try:

            # Create HDF5 file with the structure
            with h5py.File(file_name, "w") as f:

                # Set file version attribute
                f.attrs["file_version"] = "1.0"

                # Create groups
                raw_data_group = f.create_group("raw")
                paraview_group = f.create_group("paraview")
                parameters_group = f.create_group("parameters")

                # Store the filtered numpy array (with fluorophores)
                raw_data_group.create_dataset(
                    "npy", data=self.processor.filtered_numpy_array, compression="gzip"
                )

                # Store the pandas dataframe: to make sure not to depend on additional
                # dependencies (like "tables") that are not bundled with ParaView, we
                # save the dataframe as a NumPy array and save the column names and types
                # as attributes.
                self._store_dataframe(paraview_group)

                # Store important parameters
                self._store_parameters(parameters_group)

            # No error
            self._message = ""

            # Return success
            return True

        except Exception as e:

            # Set the error message
            self._message = str(e)

            return False

    @property
    def message(self):
        """Return last error message."""
        return self._message

    def _store_dataframe(self, group):
        """Write the Pandas DataFrame in a way that it can be reloaded without external dependencies."""

        dataset = group.create_dataset(
            "dataframe",
            data=self.processor.filtered_dataframe.to_numpy(),
            compression="gzip",
        )

        # Convert the column names to a list of strings
        column_names = self.processor.filtered_dataframe.columns.tolist()

        # Convert column data types to a list of strings
        column_types = [
            str(self.processor.filtered_dataframe[col].dtype)
            for col in self.processor.filtered_dataframe.columns
        ]

        # Store the column names as an attribute of the dataset
        dataset.attrs["column_names"] = column_names

        # Store column data types as attribute of the dataset
        dataset.attrs["column_types"] = column_types

        # We preserve the index as well (as a Dataset, since it can be large)
        index_data = np.array(self.processor.filtered_dataframe.index)
        group.create_dataset("dataframe_index", data=index_data, compression="gzip")

    def _store_parameters(self, group):
        """Write important parameters."""

        group.create_dataset("z_scaling_factor", data=self.processor.z_scaling_factor)
        group.create_dataset("min_trace_length", data=self.processor.min_trace_length)
        if self._state.applied_efo_thresholds is not None:
            group.create_dataset(
                "applied_efo_thresholds",
                data=self._state.applied_efo_thresholds,
            )
        if self._state.applied_cfr_thresholds is not None:
            group.create_dataset(
                "applied_cfr_thresholds",
                data=self._state.applied_cfr_thresholds,
            )
        group.create_dataset("num_fluorophores", data=self.processor.num_fluorophores)



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import pandas as pd
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QSizePolicy, QTableView

from ..processor import MinFluxProcessor
from ..state import State
from .pandas_datamodel import PandasDataModel


class DataViewer(QTableView):
    """
    A QTableWidget to display track data.
    """

    def __init__(self, *args):
        super().__init__()

        # Initialize data model to None
        self.data_model = None

        self.setSizePolicy(
            QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred
        )
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
        self.setWindowTitle("Parameters")

        # Enable sorting
        self.setSortingEnabled(True)

        # Initialize the model with an empty dataframe with the correct columns
        self.set_data(pd.DataFrame(columns=MinFluxProcessor.processed_properties()))

        # Keep a reference to the singleton State class
        self.state = State()

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            value = self.data_model[index.row()][index.column()]
            if isinstance(value, float):  # Check if the value is a float
                return f"{value:.2f}"  # Format the float value with 2 decimal places
            return value
        return None

    def set_data(self, df):
        """Display the Pandas dataframe."""
        if self.model() is not None:
            self.model().clear()
        self.data_model = PandasDataModel(df)
        self.setModel(self.data_model)

    def clear(self):
        """Clear the model."""
        if self.model() is not None:
            self.model().clear()

        # Pass an empty dataframe with correct columns
        df = pd.DataFrame(columns=MinFluxProcessor.processed_properties())
        self.set_data(df)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'analyzer.ui'
##
## Created by: Qt User Interface Compiler version 6.4.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QWidget,
)


class Ui_Analyzer(object):
    def setupUi(self, Analyzer):
        if not Analyzer.objectName():
            Analyzer.setObjectName("Analyzer")
        Analyzer.resize(1050, 800)
        Analyzer.setMinimumSize(QSize(800, 600))
        self.gridLayout = QGridLayout(Analyzer)
        self.gridLayout.setObjectName("gridLayout")
        self.tabFilterOptions = QTabWidget(Analyzer)
        self.tabFilterOptions.setObjectName("tabFilterOptions")
        sizePolicy = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.tabFilterOptions.sizePolicy().hasHeightForWidth()
        )
        self.tabFilterOptions.setSizePolicy(sizePolicy)
        self.tabFilterOptions.setMaximumSize(QSize(16777215, 16777215))
        self.tab_efo_thresholding = QWidget()
        self.tab_efo_thresholding.setObjectName("tab_efo_thresholding")
        sizePolicy1 = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(
            self.tab_efo_thresholding.sizePolicy().hasHeightForWidth()
        )
        self.tab_efo_thresholding.setSizePolicy(sizePolicy1)
        self.horizontalLayout_3 = QHBoxLayout(self.tab_efo_thresholding)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.hsEFOPeakBefore = QSpacerItem(
            20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_3.addItem(self.hsEFOPeakBefore)

        self.lbEFOExpectedCutoff = QLabel(self.tab_efo_thresholding)
        self.lbEFOExpectedCutoff.setObjectName("lbEFOExpectedCutoff")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.lbEFOExpectedCutoff.sizePolicy().hasHeightForWidth()
        )
        self.lbEFOExpectedCutoff.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.lbEFOExpectedCutoff)

        self.leEFOExpectedCutoff = QLineEdit(self.tab_efo_thresholding)
        self.leEFOExpectedCutoff.setObjectName("leEFOExpectedCutoff")
        sizePolicy2.setHeightForWidth(
            self.leEFOExpectedCutoff.sizePolicy().hasHeightForWidth()
        )
        self.leEFOExpectedCutoff.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.leEFOExpectedCutoff)

        self.pbDetectCutoffFrequency = QPushButton(self.tab_efo_thresholding)
        self.pbDetectCutoffFrequency.setObjectName("pbDetectCutoffFrequency")
        sizePolicy2.setHeightForWidth(
            self.pbDetectCutoffFrequency.sizePolicy().hasHeightForWidth()
        )
        self.pbDetectCutoffFrequency.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.pbDetectCutoffFrequency)

        self.pbEFORunFilter = QPushButton(self.tab_efo_thresholding)
        self.pbEFORunFilter.setObjectName("pbEFORunFilter")
        self.pbEFORunFilter.setEnabled(True)
        sizePolicy2.setHeightForWidth(
            self.pbEFORunFilter.sizePolicy().hasHeightForWidth()
        )
        self.pbEFORunFilter.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.pbEFORunFilter)

        self.hshsEFOPeakAfter = QSpacerItem(
            20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_3.addItem(self.hshsEFOPeakAfter)

        self.tabFilterOptions.addTab(self.tab_efo_thresholding, "")
        self.tab_cfr_thresholding = QWidget()
        self.tab_cfr_thresholding.setObjectName("tab_cfr_thresholding")
        sizePolicy1.setHeightForWidth(
            self.tab_cfr_thresholding.sizePolicy().hasHeightForWidth()
        )
        self.tab_cfr_thresholding.setSizePolicy(sizePolicy1)
        self.horizontalLayout_4 = QHBoxLayout(self.tab_cfr_thresholding)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.hsCFRFilterBefore = QSpacerItem(
            20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_4.addItem(self.hsCFRFilterBefore)

        self.checkCFRLowerThreshold = QCheckBox(self.tab_cfr_thresholding)
        self.checkCFRLowerThreshold.setObjectName("checkCFRLowerThreshold")
        sizePolicy3 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(
            self.checkCFRLowerThreshold.sizePolicy().hasHeightForWidth()
        )
        self.checkCFRLowerThreshold.setSizePolicy(sizePolicy3)

        self.horizontalLayout_4.addWidget(self.checkCFRLowerThreshold)

        self.checkCFRUpperThreshold = QCheckBox(self.tab_cfr_thresholding)
        self.checkCFRUpperThreshold.setObjectName("checkCFRUpperThreshold")
        sizePolicy3.setHeightForWidth(
            self.checkCFRUpperThreshold.sizePolicy().hasHeightForWidth()
        )
        self.checkCFRUpperThreshold.setSizePolicy(sizePolicy3)

        self.horizontalLayout_4.addWidget(self.checkCFRUpperThreshold)

        self.lbCFRFilterThreshFactor = QLabel(self.tab_cfr_thresholding)
        self.lbCFRFilterThreshFactor.setObjectName("lbCFRFilterThreshFactor")
        sizePolicy2.setHeightForWidth(
            self.lbCFRFilterThreshFactor.sizePolicy().hasHeightForWidth()
        )
        self.lbCFRFilterThreshFactor.setSizePolicy(sizePolicy2)

        self.horizontalLayout_4.addWidget(self.lbCFRFilterThreshFactor)

        self.leCFRFilterThreshFactor = QLineEdit(self.tab_cfr_thresholding)
        self.leCFRFilterThreshFactor.setObjectName("leCFRFilterThreshFactor")
        sizePolicy2.setHeightForWidth(
            self.leCFRFilterThreshFactor.sizePolicy().hasHeightForWidth()
        )
        self.leCFRFilterThreshFactor.setSizePolicy(sizePolicy2)

        self.horizontalLayout_4.addWidget(self.leCFRFilterThreshFactor)

        self.pbCFRRunAutoThreshold = QPushButton(self.tab_cfr_thresholding)
        self.pbCFRRunAutoThreshold.setObjectName("pbCFRRunAutoThreshold")
        sizePolicy2.setHeightForWidth(
            self.pbCFRRunAutoThreshold.sizePolicy().hasHeightForWidth()
        )
        self.pbCFRRunAutoThreshold.setSizePolicy(sizePolicy2)

        self.horizontalLayout_4.addWidget(self.pbCFRRunAutoThreshold)

        self.pbCFRRunFilter = QPushButton(self.tab_cfr_thresholding)
        self.pbCFRRunFilter.setObjectName("pbCFRRunFilter")
        sizePolicy2.setHeightForWidth(
            self.pbCFRRunFilter.sizePolicy().hasHeightForWidth()
        )
        self.pbCFRRunFilter.setSizePolicy(sizePolicy2)

        self.horizontalLayout_4.addWidget(self.pbCFRRunFilter)

        self.hsCFRFilterAfter = QSpacerItem(
            20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_4.addItem(self.hsCFRFilterAfter)

        self.tabFilterOptions.addTab(self.tab_cfr_thresholding, "")

        self.gridLayout.addWidget(self.tabFilterOptions, 0, 0, 1, 1)

        self.communication_layout = QHBoxLayout()
        self.communication_layout.setObjectName("communication_layout")

        self.gridLayout.addLayout(self.communication_layout, 1, 0, 1, 1)

        self.parameters_layout = QHBoxLayout()
        self.parameters_layout.setObjectName("parameters_layout")

        self.gridLayout.addLayout(self.parameters_layout, 2, 0, 1, 1)

        self.localizations_layout = QHBoxLayout()
        self.localizations_layout.setObjectName("localizations_layout")

        self.gridLayout.addLayout(self.localizations_layout, 3, 0, 1, 1)

        self.retranslateUi(Analyzer)

        self.tabFilterOptions.setCurrentIndex(0)

        QMetaObject.connectSlotsByName(Analyzer)

    # setupUi

    def retranslateUi(self, Analyzer):
        Analyzer.setWindowTitle(
            QCoreApplication.translate("Analyzer", "Analyzer", None)
        )
        self.lbEFOExpectedCutoff.setText(
            QCoreApplication.translate(
                "Analyzer", "Expected frequency of single emitters (Hz)", None
            )
        )
        self.leEFOExpectedCutoff.setText("")
        self.pbDetectCutoffFrequency.setText(
            QCoreApplication.translate("Analyzer", "Detect", None)
        )
        self.pbEFORunFilter.setText(
            QCoreApplication.translate("Analyzer", "Filter", None)
        )
        self.tabFilterOptions.setTabText(
            self.tabFilterOptions.indexOf(self.tab_efo_thresholding),
            QCoreApplication.translate("Analyzer", "EFO thresholding", None),
        )
        self.checkCFRLowerThreshold.setText(
            QCoreApplication.translate("Analyzer", "Lower bound", None)
        )
        self.checkCFRUpperThreshold.setText(
            QCoreApplication.translate("Analyzer", "Upper bound", None)
        )
        self.lbCFRFilterThreshFactor.setText(
            QCoreApplication.translate("Analyzer", "Threshold factor", None)
        )
        self.pbCFRRunAutoThreshold.setText(
            QCoreApplication.translate("Analyzer", "Threshold", None)
        )
        self.pbCFRRunFilter.setText(
            QCoreApplication.translate("Analyzer", "Filter", None)
        )
        self.tabFilterOptions.setTabText(
            self.tabFilterOptions.indexOf(self.tab_cfr_thresholding),
            QCoreApplication.translate("Analyzer", "CFR thresholding", None),
        )

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trace_stats_viewer.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QHBoxLayout,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_TraceStatsViewer(object):
    def setupUi(self, TraceStatsViewer):
        if not TraceStatsViewer.objectName():
            TraceStatsViewer.setObjectName("TraceStatsViewer")
        TraceStatsViewer.resize(850, 600)
        self.verticalLayout_2 = QVBoxLayout(TraceStatsViewer)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.main_layout = QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.commands_layout = QHBoxLayout()
        self.commands_layout.setObjectName("commands_layout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.commands_layout.addItem(self.horizontalSpacer)

        self.pbExport = QPushButton(TraceStatsViewer)
        self.pbExport.setObjectName("pbExport")

        self.commands_layout.addWidget(self.pbExport)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.commands_layout.addItem(self.horizontalSpacer_2)

        self.main_layout.addLayout(self.commands_layout)

        self.verticalLayout_2.addLayout(self.main_layout)

        self.retranslateUi(TraceStatsViewer)

        QMetaObject.connectSlotsByName(TraceStatsViewer)

    # setupUi

    def retranslateUi(self, TraceStatsViewer):
        TraceStatsViewer.setWindowTitle(
            QCoreApplication.translate("TraceStatsViewer", "Trace Stats Viewer", None)
        )
        self.pbExport.setText(
            QCoreApplication.translate("TraceStatsViewer", "Export", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trace_length_viewer.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import QApplication, QDialog, QSizePolicy, QVBoxLayout, QWidget


class Ui_TraceLengthViewer(object):
    def setupUi(self, TraceLengthViewer):
        if not TraceLengthViewer.objectName():
            TraceLengthViewer.setObjectName("TraceLengthViewer")
        TraceLengthViewer.resize(800, 600)
        self.verticalLayout_2 = QVBoxLayout(TraceLengthViewer)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.main_layout = QVBoxLayout()
        self.main_layout.setObjectName("main_layout")

        self.verticalLayout_2.addLayout(self.main_layout)

        self.retranslateUi(TraceLengthViewer)

        QMetaObject.connectSlotsByName(TraceLengthViewer)

    # setupUi

    def retranslateUi(self, TraceLengthViewer):
        TraceLengthViewer.setWindowTitle(
            QCoreApplication.translate("TraceLengthViewer", "Trace Length Viewer", None)
        )

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy.stats import mode

from pyminflux.reader import MinFluxReader


class MinFluxProcessor:
    """Processor of MINFLUX data."""

    __doc__ = """Allows for filtering and selecting data read by the underlying `MinFluxReader`. Please notice that
     `MinFluxProcessor` makes use of `State.min_trace_length` to make sure that at load and after every
      filtering step, short traces are dropped."""

    __slots__ = [
        "state",
        "reader",
        "_current_fluorophore_id",
        "_filtered_stats_dataframe",
        "_min_trace_length",
        "_selected_rows_dict",
        "_stats_to_be_recomputed",
        "_weighted_localizations",
        "_weighted_localizations_to_be_recomputed",
        "_use_weighted_localizations",
    ]

    def __init__(self, reader: MinFluxReader, min_trace_length: int = 1):
        """Constructor.

        Parameters
        ----------

        reader: MinFluxReader
            MinFluxReader object.

        min_trace_length: int (Default = 1)
            Minimum number of localizations for a trace to be kept. Shorter traces are dropped.
        """

        # Store a reference to the MinFluxReader
        self.reader: MinFluxReader = reader

        # Global options
        self._min_trace_length: int = min_trace_length

        # Cache the filtered stats dataframe
        self._filtered_stats_dataframe = None

        # Keep separate arrays of booleans to cache selection state for all fluorophores IDs.
        self._selected_rows_dict = None
        self._init_selected_rows_dict()

        # Keep track of the selected fluorophore
        # 0 - All (default)
        # 1 - Fluorophore 1
        # 2 - Fluorophore 2
        self._current_fluorophore_id = 0

        # Cache the weighted, averaged TID positions
        self._weighted_localizations = None

        # Keep track whether the statistics and the weighted localizations need to be recomputed
        self._stats_to_be_recomputed = False
        self._weighted_localizations_to_be_recomputed = False

        # Whether to use weighted average for localizations
        self._use_weighted_localizations = False

        # Apply the global filters
        self._apply_global_filters()

    def _init_selected_rows_dict(self):
        """Initialize the selected rows array."""
        # How many fluorophores do we have?
        self._selected_rows_dict = {
            1: pd.Series(
                data=np.ones(len(self.full_dataframe.index), dtype=bool),
                index=self.full_dataframe.index,
            ),
            2: pd.Series(
                data=np.ones(len(self.full_dataframe.index), dtype=bool),
                index=self.full_dataframe.index,
            ),
        }

    @property
    def min_trace_length(self):
        """Minimum number of localizations for the trace to be kept."""
        return self._min_trace_length

    @property
    def z_scaling_factor(self):
        """Returns the scaling factor for the z coordinates from the underlying MinFluxReader."""
        return self.reader.z_scaling_factor

    @min_trace_length.setter
    def min_trace_length(self, value):
        if value < 1 or int(value) != value:
            raise ValueError(
                "MinFluxProcessor.min_trace_length must be a positive integer!"
            )
        self._min_trace_length = value

    @property
    def is_3d(self) -> bool:
        """Return True if the acquisition is 3D.

        Returns
        -------

        is_3d: bool
            True if the acquisition is 3D, False otherwise.
        """
        return self.reader.is_3d

    @property
    def num_values(self) -> int:
        """Return the number of values in the (filtered) dataframe.

        Returns
        -------

        n: int
            Number of values in the dataframe after all filters have been applied.
        """
        if self.filtered_dataframe is not None:
            return len(self.filtered_dataframe.index)

        return 0

    @property
    def current_fluorophore_id(self) -> int:
        """Return current fluorophore ID (0 for all)."""
        return self._current_fluorophore_id

    @current_fluorophore_id.setter
    def current_fluorophore_id(self, fluorophore_id: int) -> None:
        """Set current fluorophore ID (0 for all)."""

        if fluorophore_id not in [0, 1, 2]:
            raise ValueError(f"Only 1 or 2 are valid fluorophore IDs.")

        # Set the new fluorophore_id
        self._current_fluorophore_id = fluorophore_id

        # Apply the global filters
        self._apply_global_filters()

        # Flag stats to be recomputed
        self._stats_to_be_recomputed = True

    @property
    def num_fluorophores(self) -> int:
        """Return the number of fluorophores."""
        return len(np.unique(self.full_dataframe["fluo"].values))

    @property
    def filtered_numpy_array_all(self):
        """Return the raw NumPy array with applied filters (for all fluorophores)."""

        # Copy the raw NumPy array
        raw_array = self.reader.valid_raw_data
        if raw_array is None:
            return None

        # Append the fluorophore ID data
        raw_array["fluo"] = self.full_dataframe["fluo"].astype(np.uint8)

        # Extract combination of fluorophore 1 and 2 filtered dataframes
        mask_1 = (self.full_dataframe["fluo"] == 1) & self._selected_rows_dict[1]
        mask_2 = (self.full_dataframe["fluo"] == 2) & self._selected_rows_dict[2]
        return raw_array[mask_1 | mask_2]

    @property
    def filtered_numpy_array(self):
        """Return the raw NumPy array with applied filters for the selected fluorophores."""

        # Copy the raw NumPy array
        full_array = self.filtered_numpy_array_all
        if full_array is None:
            return None

        if self.current_fluorophore_id == 0:
            return full_array
        elif self.current_fluorophore_id == 1:
            return full_array[full_array["fluo"] == 1]
        elif self.current_fluorophore_id == 2:
            return full_array[full_array["fluo"] == 2]
        else:
            raise ValueError(
                f"Unexpected fluorophore ID {self.current_fluorophore_id}."
            )

    @property
    def full_dataframe(self) -> Union[None, pd.DataFrame]:
        """Return the full dataframe (with valid entries only), with no selections or filters.

        Returns
        -------

        full_dataframe: Union[None, pd.DataFrame]
            A Pandas dataframe or None if no file was loaded.
        """
        return self.reader.processed_dataframe

    @property
    def filename(self) -> Union[Path, None]:
        """Return the filename if set."""
        if self.reader is None:
            return None
        return self.reader.filename

    def _filtered_dataframe_all(self) -> Union[None, pd.DataFrame]:
        """Return joint dataframe for all fluorophores and with all filters applied.

        Returns
        -------

        filtered_dataframe_all: Union[None, pd.DataFrame]
            A Pandas dataframe or None if no file was loaded.
        """
        if self.full_dataframe is None:
            return None

        # Extract combination of fluorophore 1 and 2 filtered dataframes
        mask_1 = (self.full_dataframe["fluo"] == 1) & self._selected_rows_dict[1]
        mask_2 = (self.full_dataframe["fluo"] == 2) & self._selected_rows_dict[2]
        return self.full_dataframe.loc[mask_1 | mask_2]

    @property
    def filtered_dataframe(self) -> Union[None, pd.DataFrame]:
        """Return dataframe with all filters applied.

        Returns
        -------

        filtered_dataframe: Union[None, pd.DataFrame]
            A Pandas dataframe or None if no file was loaded.
        """
        if self.full_dataframe is None:
            return None
        if self.current_fluorophore_id == 0:
            return self._filtered_dataframe_all()
        else:
            df = self.full_dataframe.loc[
                self.full_dataframe["fluo"] == self.current_fluorophore_id
            ]
            return df.loc[self._selected_rows_dict[self.current_fluorophore_id]]

    @property
    def filtered_dataframe_stats(self) -> Union[None, pd.DataFrame]:
        """Return dataframe stats with all filters applied.

        Returns
        -------

        filtered_dataframe_stats: Union[None, pd.DataFrame]
            A Pandas dataframe with all data statistics or None if no file was loaded.
        """
        if self._stats_to_be_recomputed:
            self._calculate_statistics()
        return self._filtered_stats_dataframe

    @property
    def weighted_localizations(self) -> Union[None, pd.DataFrame]:
        """Return the average (x, y, z) position per TID weighted by the relative photon count."""
        if self._weighted_localizations_to_be_recomputed:
            self._calculate_weighted_positions()
        return self._weighted_localizations

    @property
    def use_weighted_localizations(self) -> bool:
        """Whether to use weighted average to calculate the mean localization per TID."""
        return self._use_weighted_localizations

    @use_weighted_localizations.setter
    def use_weighted_localizations(self, value: bool):
        """Whether to use weighted average to calculate the mean localization per TID."""
        self._use_weighted_localizations = value
        self._weighted_localizations_to_be_recomputed = True

    @classmethod
    def processed_properties(cls):
        """Return the processed dataframe columns."""
        return MinFluxReader.processed_properties()

    @classmethod
    def trace_stats_properties(cls):
        """Return the columns of the filtered_dataframe_stats."""
        return [
            "tid",
            "n",
            "mx",
            "my",
            "mz",
            "sx",
            "sy",
            "sxy",
            "exy",
            "rms_xy",
            "sz",
            "ez",
            "fluo",
        ]

    def reset(self):
        """Drops all dynamic filters and resets the data to the processed data frame with global filters."""

        # Clear the selection per fluorophore; they will be reinitialized as
        # all selected at the first access.
        self._init_selected_rows_dict()

        # Reset the mapping to the corresponding fluorophore
        self.full_dataframe["fluo"] = 1

        # Default fluorophore is 0 (no selection)
        self.current_fluorophore_id = 0

        # Apply global filters
        self._apply_global_filters()

    def update_localizations(
        self, x: np.ndarray, y: np.ndarray, z: Optional[np.ndarray] = None
    ):
        """Updates the localization coordinates of current filtered dataframe.

        This can be used for instance after a drift correction.

        Parameters
        ----------

        x: np.ndarray
            Array of x coordinates.

        y: np.ndarray
            Array of y coordinates.

        z: np.ndarray (Optional)
            Optional array of z coordinates. Omit it to skip.
            If the acquisition is 2D, it will be ignored in any case.
        """

        if self.full_dataframe is None:
            return

        # Make sure to work with NumPy arrays
        x = np.array(x)
        y = np.array(y)
        if z is not None and self.is_3d:
            z = np.array(z)

        # Select the correct rows to update
        if self.current_fluorophore_id == 0:
            mask_1 = (self.full_dataframe["fluo"] == 1) & self._selected_rows_dict[1]
            mask_2 = (self.full_dataframe["fluo"] == 2) & self._selected_rows_dict[2]
            mask = mask_1 | mask_2
        elif self.current_fluorophore_id == 1:
            mask = (self.full_dataframe["fluo"] == 1) & self._selected_rows_dict[1]
        else:
            mask = (self.full_dataframe["fluo"] == 2) & self._selected_rows_dict[2]

        # Make sure that the lengths match
        assert np.sum(mask.values) == len(x), "Unexpected number of elements in x."
        assert np.sum(mask.values) == len(y), "Unexpected number of elements in y."
        if z is not None and self.is_3d:
            assert np.sum(mask.values) == len(z), "Unexpected number of elements in z."

        # Re-assign the data at the reader level
        self.reader._data_df.loc[mask, "x"] = x
        self.reader._data_df.loc[mask, "y"] = y
        if z is not None and self.is_3d:
            self.reader._data_df.loc[mask, "z"] = z

        # Also update the raw structured NumPy array. Since NumPy
        # will return a copy if we try to access the "loc" array
        # directly using logical arrays, we need to iterate over
        # all rows one by one!
        #
        # Furthermore, we need to scale the values by the factor
        # self.reader._unit_scaling_factor
        x_scaled = x / self.reader._unit_scaling_factor
        y_scaled = y / self.reader._unit_scaling_factor
        if z is not None and self.is_3d:
            z_scaled = z / self.reader._unit_scaling_factor
        idx = self.reader._loc_index
        vld_indices = np.where(self.reader._valid_entries)[0]
        mask_indices = np.where(mask)[0]
        for i, I in enumerate(mask_indices):
            if I in vld_indices:
                self.reader._data_array[I]["itr"][idx]["loc"][0] = x_scaled[i]
                self.reader._data_array[I]["itr"][idx]["loc"][1] = y_scaled[i]
                if z is not None and self.is_3d:
                    self.reader._data_array[I]["itr"][idx]["loc"][2] = z_scaled[i]

        # Mark derived data to be recomputed
        self._stats_to_be_recomputed = True
        self._weighted_localizations_to_be_recomputed = True

    def set_fluorophore_ids(self, fluorophore_ids: np.ndarray[int]):
        """Assign the fluorophore IDs."""
        if len(fluorophore_ids) != len(self.full_dataframe.index):
            raise ValueError(
                "The number of fluorophore IDs does not match the number of entries in the dataframe."
            )
        self.full_dataframe["fluo"] = fluorophore_ids
        self._init_selected_rows_dict()
        self._apply_global_filters()

    def select_by_indices(
        self, indices, from_weighted_locs: bool = False
    ) -> Union[None, pd.DataFrame]:
        """Return view on a subset of the filtered dataset or the weighted localisations defined by the passed indices.

        The underlying dataframe is not modified.

        Parameters
        ----------

        indices: array
            Logical array for selecting the elements to be returned.

        from_weighted_locs: bool
            If True, select from the weighted_localizations dataframe; otherwise, from the filtered_dataframe.

        Returns
        -------

        subset: Union[None, pd.DataFrame]
            A view on a subset of the dataframe defined by the passed indices, or None if no file was loaded.
        """
        if from_weighted_locs:
            if self._weighted_localizations is None:
                return None
            return self._weighted_localizations.iloc[indices]
        else:
            if self.filtered_dataframe is None:
                return None
            return self.filtered_dataframe.iloc[indices]

    def select_by_1d_range(
        self, x_prop, x_range, from_weighted_locs: bool = False
    ) -> Union[None, pd.DataFrame]:
        """Return a view on a subset of the filtered dataset or the weighted localisations defined by the passed
        parameter and corresponding range.

        The underlying dataframe is not modified.

        Parameters
        ----------

        x_prop: str
            Property to be filtered by corresponding x_range.

        x_range: tuple
            Tuple containing the minimum and maximum values for the selected property.

        from_weighted_locs: bool
            If True, select from the weighted_localizations dataframe; otherwise, from the filtered_dataframe.

        Returns
        -------

        subset: Union[None, pd.DataFrame]
            A view on a subset of the dataframe defined by the passed property and range, or None if no file was loaded.
        """

        # Make sure that the range is increasing
        x_min = x_range[0]
        x_max = x_range[1]
        if x_max < x_min:
            x_max, x_min = x_min, x_max

        if from_weighted_locs:
            return self._weighted_localizations.loc[
                (self._weighted_localizations[x_prop] >= x_min)
                & (self._weighted_localizations[x_prop] < x_max)
            ]
        else:
            # Work with currently selected rows
            df = self.filtered_dataframe

            return df.loc[(df[x_prop] >= x_min) & (df[x_prop] < x_max)]

    def select_by_2d_range(
        self, x_prop, y_prop, x_range, y_range, from_weighted_locs: bool = False
    ) -> Union[None, pd.DataFrame]:
        """Return a view on a subset of the filtered dataset or the weighted localisations defined by the passed
        parameters and corresponding ranges.

        The underlying dataframe is not modified.

        Parameters
        ----------

        x_prop: str
            First property to be filtered by corresponding x_range.

        y_prop: str
            Second property to be filtered by corresponding y_range.

        x_range: tuple
            Tuple containing the minimum and maximum values for the first property.

        y_range: tuple
            Tuple containing the minimum and maximum values for the second property.

        from_weighted_locs: bool
            If True, select from the weighted_localizations dataframe; otherwise, from the filtered_dataframe.

        Returns
        -------

        subset: Union[None, pd.DataFrame]
            A view on a subset of the dataframe defined by the passed properties and ranges, or None if no file
            was loaded.
        """

        # Make sure that the ranges are increasing
        x_min = x_range[0]
        x_max = x_range[1]
        if x_max < x_min:
            x_max, x_min = x_min, x_max

        y_min = y_range[0]
        y_max = y_range[1]
        if y_max < y_min:
            y_max, y_min = y_min, y_max

        if from_weighted_locs:
            return self._weighted_localizations.loc[
                (self._weighted_localizations[x_prop] >= x_min)
                & (self._weighted_localizations[x_prop] < x_max)
                & (self._weighted_localizations[y_prop] >= y_min)
                & (self._weighted_localizations[y_prop] < y_max)
            ]
        else:
            # Work with currently selected rows
            df = self.filtered_dataframe

            return df.loc[
                (df[x_prop] >= x_min)
                & (df[x_prop] < x_max)
                & (df[y_prop] >= y_min)
                & (df[y_prop] < y_max)
            ]

    def filter_by_2d_range(self, x_prop, y_prop, x_range, y_range):
        """Filter dataset by the extracting a rectangular ROI over two parameters and two ranges.

        Parameters
        ----------

        x_prop: str
            First property to be filtered by corresponding x_range.

        y_prop: str
            Second property to be filtered by corresponding y_range.

        x_range: tuple
            Tuple containing the minimum and maximum values for the first property.

        y_range: tuple
            Tuple containing the minimum and maximum values for the second property.
        """

        # Make sure that the ranges are increasing
        x_min = x_range[0]
        x_max = x_range[1]
        if x_max < x_min:
            x_max, x_min = x_min, x_max

        y_min = y_range[0]
        y_max = y_range[1]
        if y_max < y_min:
            y_max, y_min = y_min, y_max

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 1:
            self._selected_rows_dict[1] = (
                self._selected_rows_dict[1]
                & (self.filtered_dataframe[x_prop] >= x_min)
                & (self.filtered_dataframe[x_prop] < x_max)
                & (self.filtered_dataframe[y_prop] >= y_min)
                & (self.filtered_dataframe[y_prop] < y_max)
            )

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 2:
            self._selected_rows_dict[2] = (
                self._selected_rows_dict[2]
                & (self.filtered_dataframe[x_prop] >= x_min)
                & (self.filtered_dataframe[x_prop] < x_max)
                & (self.filtered_dataframe[y_prop] >= y_min)
                & (self.filtered_dataframe[y_prop] < y_max)
            )

        # Make sure to always apply the global filters
        self._apply_global_filters()

        # Make sure to flag the derived data to be recomputed
        self._stats_to_be_recomputed = True
        self._weighted_localizations_to_be_recomputed = True

    def _apply_global_filters(self):
        """Apply filters that are defined in the global application configuration."""

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 1:
            self._selected_rows_dict[1] = self._filter_by_tid_length(1)

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 2:
            self._selected_rows_dict[2] = self._filter_by_tid_length(2)

        # Make sure to flag the derived data to be recomputed
        self._stats_to_be_recomputed = True
        self._weighted_localizations_to_be_recomputed = True

    def _filter_by_tid_length(self, index):
        # Make sure to count only currently selected rows
        df = self.full_dataframe.copy()
        df.loc[np.invert(self._selected_rows_dict[index]), "tid"] = np.nan

        # Select all rows where the count of TIDs is larger than self._min_trace_num
        counts = df["tid"].value_counts(normalize=False)
        return df["tid"].isin(counts[counts >= self.min_trace_length].index)

    def filter_by_single_threshold(
        self, prop: str, threshold: Union[int, float], larger_than: bool = True
    ):
        """Apply single threshold to filter values either lower or higher (equal) than threshold for given property."""

        # Apply filter
        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 1:
            if larger_than:
                self._selected_rows_dict[1] = self._selected_rows_dict[1] & (
                    self.filtered_dataframe[prop] >= threshold
                )
            else:
                self._selected_rows_dict[1] = self._selected_rows_dict[1] & (
                    self.filtered_dataframe[prop] < threshold
                )

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 2:
            if larger_than:
                self._selected_rows_dict[2] = self._selected_rows_dict[2] & (
                    self.filtered_dataframe[prop] >= threshold
                )
            else:
                self._selected_rows_dict[2] = self._selected_rows_dict[2] & (
                    self.filtered_dataframe[prop] < threshold
                )

        # Apply the global filters
        self._apply_global_filters()

        # Make sure to flag the derived data to be recomputed
        self._stats_to_be_recomputed = True
        self._weighted_localizations_to_be_recomputed = True

    def filter_by_1d_range(self, x_prop: str, x_range: tuple):
        """Apply min and max thresholding to the given property.

        Parameters
        ----------

        x_prop: str
            Name of the property (dataframe column) to filter.

        x_range: tuple
            Tuple containing the minimum and maximum values for the selected property.
        """

        # Make sure that the ranges are increasing
        x_min = x_range[0]
        x_max = x_range[1]
        if x_max < x_min:
            x_max, x_min = x_min, x_max

        # Apply filter
        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 1:
            self._selected_rows_dict[1] = (
                self._selected_rows_dict[1]
                & (self.filtered_dataframe[x_prop] >= x_min)
                & (self.filtered_dataframe[x_prop] < x_max)
            )

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 2:
            self._selected_rows_dict[2] = (
                self._selected_rows_dict[2]
                & (self.filtered_dataframe[x_prop] >= x_min)
                & (self.filtered_dataframe[x_prop] < x_max)
            )

        # Apply the global filters
        self._apply_global_filters()

        # Make sure to flag the derived data to be recomputed
        self._stats_to_be_recomputed = True
        self._weighted_localizations_to_be_recomputed = True

    def filter_by_1d_range_complement(self, prop: str, x_range: tuple):
        """Apply min and max thresholding to the given property but keep the
        data outside the range (i.e., crop the selected range).

        Parameters
        ----------

        prop: str
            Name of the property (dataframe column) to filter.

        x_range: tuple
            Tuple containing the minimum and maximum values for cropping the selected property.
        """

        # Make sure that the ranges are increasing
        x_min = x_range[0]
        x_max = x_range[1]
        if x_max < x_min:
            x_max, x_min = x_min, x_max

        # Apply filter
        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 1:
            self._selected_rows_dict[1] = self._selected_rows_dict[1] & (
                (self.filtered_dataframe[prop] < x_min)
                | (self.filtered_dataframe[prop] >= x_max)
            )

        if self.current_fluorophore_id == 0 or self.current_fluorophore_id == 2:
            self._selected_rows_dict[2] = self._selected_rows_dict[2] & (
                (self.filtered_dataframe[prop] < x_min)
                | (self.filtered_dataframe[prop] >= x_max)
            )

        # Apply the global filters
        self._apply_global_filters()

        # Make sure to flag the derived data to be recomputed
        self._stats_to_be_recomputed = True
        self._weighted_localizations_to_be_recomputed = True

    def _calculate_statistics(self):
        """Calculate per-trace statistics."""

        # Make sure we have processed dataframe to work on
        if self.full_dataframe is None:
            return

        # Only recompute statistics if needed
        if not self._stats_to_be_recomputed:
            return

        # Work with currently selected rows
        df = self.filtered_dataframe

        # Calculate the statistics
        df_tid = self.calculate_statistics_on(df)

        # Store the results
        self._filtered_stats_dataframe = df_tid

        # Flag the statistics to be computed
        self._stats_to_be_recomputed = False

    @staticmethod
    def calculate_statistics_on(df: pd.DataFrame) -> pd.DataFrame:
        """Calculate per-trace statistics for the selected dataframe.

        Parameters
        ----------

        df: pd.DataFrame
            DataFrame (view) generated by one of the `select_by_*` methods.

        Returns
        -------
        df_stats: pd.DataFrame
            Per-trace statistics calculated on the passed DataFrame selection (view).
        """

        # Calculate some statistics per TID on the passed dataframe
        df_grouped = df.groupby("tid")

        tid = df_grouped["tid"].first().values
        n = df_grouped["tid"].count().values
        mx = df_grouped["x"].mean().values
        my = df_grouped["y"].mean().values
        mz = df_grouped["z"].mean().values
        sx = df_grouped["x"].std().values
        sy = df_grouped["y"].std().values
        sz = df_grouped["z"].std().values
        tmp = np.power(sx, 2) + np.power(sy, 2)
        sxy = np.sqrt(tmp)
        rms_xy = np.sqrt(tmp / 2)
        exy = sxy / np.sqrt(n)
        ez = sz / np.sqrt(n)
        fluo = df_grouped["fluo"].agg(lambda x: mode(x, keepdims=True)[0][0]).values

        # Prepare a dataframe with the statistics
        df_tid = pd.DataFrame(columns=MinFluxProcessor.trace_stats_properties())

        # Store trace stats
        df_tid["tid"] = tid  # Trace ID
        df_tid["n"] = n  # Number of localizations for given trace ID
        df_tid["mx"] = mx  # x mean localization
        df_tid["my"] = my  # y mean localization
        df_tid["mz"] = mz  # z mean localization
        df_tid["sx"] = sx  # x localization precision
        df_tid["sy"] = sy  # y localization precision
        df_tid["sxy"] = sxy  # Lateral (x, y) localization precision
        df_tid["rms_xy"] = rms_xy  # Lateral root mean square
        df_tid["exy"] = exy  # Standard error of sxy
        df_tid["sz"] = sz  # z localization precision
        df_tid["ez"] = ez  # Standard error of ez
        df_tid["fluo"] = fluo  # Assigned fluorophore ID

        # ["sx", "sy", "sxy", "rms_xy", "exy", "sz", "ez"] columns will contain
        # np.nan if n == 1: we replace them with 0.0.
        # @TODO: should this be changed? It could be a global option.
        df_tid[["sx", "sy", "sxy", "rms_xy", "exy", "sz", "ez"]] = df_tid[
            ["sx", "sy", "sxy", "rms_xy", "exy", "sz", "ez"]
        ].fillna(value=0.0)

        # Return the results
        return df_tid

    def _calculate_weighted_positions(self):
        """Calculate per-trace localization weighted by relative photon count."""

        # Make sure we have processed dataframe to work on
        if self.full_dataframe is None:
            return

        # Only recompute weighted localizations if needed
        if not self._weighted_localizations_to_be_recomputed:
            return

        # Work with currently selected rows
        df = self.filtered_dataframe

        # Normal or weighted averaging?
        if self._use_weighted_localizations:
            # Calculate weighing factors for TIDs
            df = df.assign(
                eco_rel=df["eco"].groupby(df["tid"]).transform(lambda x: x / x.sum())
            )

            # Calculate relative contributions of (x, y, z)_i for each TID
            df["x_rel"] = df["x"] * df["eco_rel"]
            df["y_rel"] = df["y"] * df["eco_rel"]
            df["z_rel"] = df["z"] * df["eco_rel"]

            # Calculate the weighted localizations
            df_grouped = df.groupby("tid")
            tid = df_grouped["tid"].first().values
            x_w = df_grouped["x_rel"].sum().values
            y_w = df_grouped["y_rel"].sum().values
            z_w = df_grouped["z_rel"].sum().values
            fluo = (
                df_grouped["fluo"].agg(lambda x: mode(x, keepdims=True)[0][0]).values
            )  # Return the most frequent fluo ID

        else:
            # Calculate simple average of localizations by TID
            df_grouped = df.groupby(df["tid"])
            tid = df_grouped["tid"].first().values
            x_w = df_grouped["x"].mean().values
            y_w = df_grouped["y"].mean().values
            z_w = df_grouped["z"].mean().values
            fluo = (
                df_grouped["fluo"].agg(lambda x: mode(x, keepdims=True)[0][0]).values
            )  # Return the most frequent fluo ID

        # Prepare a dataframe with the weighted localizations
        df_loc = pd.DataFrame(columns=["tid", "x", "y", "z", "fluo"])
        df_loc["tid"] = tid
        df_loc["x"] = x_w
        df_loc["y"] = y_w
        df_loc["z"] = z_w
        df_loc["fluo"] = fluo

        # Store the results
        self._weighted_localizations = df_loc

        # Make sure to flag the derived data to be recomputed
        self._weighted_localizations_to_be_recomputed = False



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
import platform
from pathlib import Path

from PySide6.QtCore import QSettings

from pyminflux import __APP_NAME__


class Settings:
    def __init__(self):
        """Constructor."""

        # Define the organization and application names
        self.org_name = "ch.ethz"
        self.app_name = __APP_NAME__

        # Construct the file path for the INI file based on the platform
        if platform.system() == "Windows":
            config_folder = Path(os.environ["APPDATA"]) / self.org_name
        else:
            config_folder = Path.home() / ".config" / self.org_name
        config_folder.mkdir(parents=True, exist_ok=True)
        self.config_file_path = config_folder / f"{self.app_name}.ini"

        # Create the QSettings object
        self._settings = QSettings(str(self.config_file_path), QSettings.IniFormat)

    @property
    def instance(self):
        """Give access to the underlying QSettings object."""
        return self._settings



# -*- coding: utf-8 -*-

#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

################################################################################
## Form generated from reading UI file 'plotter_toolbar.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)


class Ui_PlotterToolbar(object):
    def setupUi(self, PlotterToolbar):
        if not PlotterToolbar.objectName():
            PlotterToolbar.setObjectName("PlotterToolbar")
        PlotterToolbar.resize(1077, 34)
        sizePolicy = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(PlotterToolbar.sizePolicy().hasHeightForWidth())
        PlotterToolbar.setSizePolicy(sizePolicy)
        self.horizontalLayout = QHBoxLayout(PlotterToolbar)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalLayout.setContentsMargins(-1, 3, -1, 3)
        self.horizontalSpacer_2 = QSpacerItem(
            10, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.lbOptions = QLabel(PlotterToolbar)
        self.lbOptions.setObjectName("lbOptions")

        self.horizontalLayout.addWidget(self.lbOptions)

        self.cbColorCodeSelector = QComboBox(PlotterToolbar)
        self.cbColorCodeSelector.addItem("")
        self.cbColorCodeSelector.addItem("")
        self.cbColorCodeSelector.addItem("")
        self.cbColorCodeSelector.setObjectName("cbColorCodeSelector")

        self.horizontalLayout.addWidget(self.cbColorCodeSelector)

        self.cbPlotAveragePos = QCheckBox(PlotterToolbar)
        self.cbPlotAveragePos.setObjectName("cbPlotAveragePos")

        self.horizontalLayout.addWidget(self.cbPlotAveragePos)

        self.line = QFrame(PlotterToolbar)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.VLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.horizontalLayout.addWidget(self.line)

        self.cbFirstParam = QComboBox(PlotterToolbar)
        self.cbFirstParam.setObjectName("cbFirstParam")
        sizePolicy.setHeightForWidth(self.cbFirstParam.sizePolicy().hasHeightForWidth())
        self.cbFirstParam.setSizePolicy(sizePolicy)

        self.horizontalLayout.addWidget(self.cbFirstParam)

        self.cbSecondParam = QComboBox(PlotterToolbar)
        self.cbSecondParam.setObjectName("cbSecondParam")
        sizePolicy.setHeightForWidth(
            self.cbSecondParam.sizePolicy().hasHeightForWidth()
        )
        self.cbSecondParam.setSizePolicy(sizePolicy)

        self.horizontalLayout.addWidget(self.cbSecondParam)

        self.pbPlot = QPushButton(PlotterToolbar)
        self.pbPlot.setObjectName("pbPlot")
        sizePolicy.setHeightForWidth(self.pbPlot.sizePolicy().hasHeightForWidth())
        self.pbPlot.setSizePolicy(sizePolicy)

        self.horizontalLayout.addWidget(self.pbPlot)

        self.horizontalSpacer_4 = QSpacerItem(
            10, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_4)

        self.retranslateUi(PlotterToolbar)

        QMetaObject.connectSlotsByName(PlotterToolbar)

    # setupUi

    def retranslateUi(self, PlotterToolbar):
        PlotterToolbar.setWindowTitle(
            QCoreApplication.translate("PlotterToolbar", "Form", None)
        )
        self.lbOptions.setText(
            QCoreApplication.translate("PlotterToolbar", "Options:", None)
        )
        self.cbColorCodeSelector.setItemText(
            0, QCoreApplication.translate("PlotterToolbar", "No color code", None)
        )
        self.cbColorCodeSelector.setItemText(
            1, QCoreApplication.translate("PlotterToolbar", "Color-code by TID", None)
        )
        self.cbColorCodeSelector.setItemText(
            2,
            QCoreApplication.translate(
                "PlotterToolbar", "Color-code by fluorophore", None
            ),
        )

        self.cbPlotAveragePos.setText(
            QCoreApplication.translate("PlotterToolbar", "Avg Loc (TID)", None)
        )
        self.pbPlot.setText(QCoreApplication.translate("PlotterToolbar", "Plot", None))

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import sys
from datetime import datetime
from pathlib import Path

from pyqtgraph import ViewBox
from PySide6 import QtGui
from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import QDesktopServices, QIcon
from PySide6.QtWidgets import (
    QDialog,
    QDockWidget,
    QFileDialog,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QTextBrowser,
    QTextEdit,
    QVBoxLayout,
)

import pyminflux.resources
from pyminflux import __APP_NAME__, __version__
from pyminflux.processor import MinFluxProcessor
from pyminflux.reader import MinFluxReader, NativeMetadataReader
from pyminflux.settings import Settings
from pyminflux.state import State
from pyminflux.ui.analyzer import Analyzer
from pyminflux.ui.color_unmixer import ColorUnmixer
from pyminflux.ui.dataviewer import DataViewer
from pyminflux.ui.emittingstream import EmittingStream
from pyminflux.ui.frc_tool import FRCTool
from pyminflux.ui.options import Options
from pyminflux.ui.plotter import Plotter
from pyminflux.ui.plotter_toolbar import PlotterToolbar
from pyminflux.ui.time_inspector import TimeInspector
from pyminflux.ui.trace_length_viewer import TraceLengthViewer
from pyminflux.ui.trace_stats_viewer import TraceStatsViewer
from pyminflux.ui.ui_main_window import Ui_MainWindow
from pyminflux.ui.wizard import WizardDialog
from pyminflux.utils import check_for_updates
from pyminflux.writer import MinFluxWriter, PyMinFluxNativeWriter


class PyMinFluxMainWindow(QMainWindow, Ui_MainWindow):
    """
    Main application window.
    """

    request_sync_external_tools = Signal(None, name="request_sync_external_tools")

    def __init__(self, parent=None):
        """
        Constructor.
        """

        # Call the base constructor
        super().__init__(parent)

        # Initialize the dialog
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # Main window title
        self.setWindowTitle(f"{__APP_NAME__} v{__version__}")

        # Set the window icon
        icon = QIcon(":/icons/icon.png")
        self.setWindowIcon(icon)

        # Keep a reference to the state machine
        self.state = State()

        # Keep track of the last selected path
        self.last_selected_path = ""

        # Keep a reference to the MinFluxProcessor
        self.processor = None

        # Read the application settings and update the state
        self.load_and_apply_settings()

        # Set the menu state based on the settings
        self.ui.actionConsole.setChecked(self.state.open_console_at_start)

        # Dialogs and widgets
        self.data_viewer = None
        self.analyzer = None
        self.plotter = None
        self.color_unmixer = None
        self.time_inspector = None
        self.trace_stats_viewer = None
        self.trace_length_viewer = None
        self.frc_tool = None
        self.options = Options()

        # Initialize widget and its dock
        self.wizard = WizardDialog()
        self.wizard_dock = QDockWidget("", self)
        self.wizard_dock.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        self.wizard_dock.setFeatures(
            QDockWidget.DockWidgetMovable | QDockWidget.DockWidgetFloatable
        )
        self.wizard_dock.setWidget(self.wizard)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.wizard_dock)

        # Initialize Plotter and DataViewer
        self.plotter = Plotter()
        self.plotter_toolbar = PlotterToolbar()
        self.data_viewer = DataViewer()

        # Create the output console
        self.txConsole = QTextEdit()
        self.txConsole.setReadOnly(True)
        self.txConsole.setMinimumHeight(100)
        self.txConsole.setMaximumHeight(500)
        self.txConsole.setLineWrapMode(QTextEdit.NoWrap)
        self.txConsole.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        # Add them to the splitter
        self.ui.splitter_layout.addWidget(self.plotter)
        self.ui.splitter_layout.addWidget(self.plotter_toolbar)
        self.ui.splitter_layout.addWidget(self.data_viewer)
        self.ui.splitter_layout.addWidget(self.txConsole)

        # Make sure to only show the console if requested
        self.toggle_console_visibility()

        # Set initial visibility and enabled states
        self.enable_ui_components(False)

        # Set up signals and slots
        self.setup_conn()

        # Install the custom output stream
        sys.stdout = EmittingStream()
        sys.stdout.signal_textWritten.connect(self.print_to_console)

        # Print a welcome message to the console
        print(f"Welcome to {__APP_NAME__}.")

    def load_and_apply_settings(self):
        """Read the application settings and update the State."""

        # Open settings file
        settings = Settings()

        # Read and set 'last_selected_path' option
        self.last_selected_path = Path(
            settings.instance.value("io/last_selected_path", ".")
        )

        # Read and set 'min_trace_length' option
        self.state.min_trace_length = int(
            settings.instance.value(
                "options/min_trace_length", self.state.min_trace_length
            )
        )

        # @SUPERSEDED: Read 'min_trace_length' and set as 'min_trace_length'
        # This property will be removed from the file the next time the user hits
        # the "Set as new default" in the Options dialog.
        if settings.instance.contains("options/min_trace_length"):
            self.state.min_trace_length = int(
                settings.instance.value(
                    "options/min_trace_length", self.state.min_trace_length
                )
            )

        # Current option name: "options/min_trace_length"
        if settings.instance.contains("options/min_trace_length"):
            self.state.min_trace_length = int(
                settings.instance.value(
                    "options/min_trace_length", self.state.min_trace_length
                )
            )

        # Read and set 'z_scaling_factor' option
        self.state.z_scaling_factor = float(
            settings.instance.value(
                "options/z_scaling_factor", self.state.z_scaling_factor
            )
        )

        # Read and set 'efo_bin_size_hz' option
        self.state.efo_bin_size_hz = float(
            settings.instance.value(
                "options/efo_bin_size_hz", self.state.efo_bin_size_hz
            )
        )

        # Read and set 'efo_expected_frequency' option
        self.state.efo_expected_frequency = float(
            settings.instance.value(
                "options/efo_expected_frequency", self.state.efo_expected_frequency
            )
        )

        # Read and set 'weigh_avg_localization_by_eco' option
        value = settings.instance.value(
            "options/weigh_avg_localization_by_eco",
            self.state.weigh_avg_localization_by_eco,
        )
        weigh_avg_localization_by_eco = (
            value.lower() == "true" if isinstance(value, str) else bool(value)
        )
        self.state.weigh_avg_localization_by_eco = weigh_avg_localization_by_eco

        # Read and set the plot ranges
        value = settings.instance.value("options/efo_range", self.state.efo_range)
        if value == "None" or value is None:
            self.state.efo_range = None
        elif type(value) is list:
            self.state.efo_range = (float(value[0]), float(value[1]))
        else:
            raise ValueError("Unexpected value for 'efo_range' in settings.")

        value = settings.instance.value("options/cfr_range", self.state.cfr_range)
        if value == "None" or value is None:
            self.state.cfr_range = None
        elif type(value) is list:
            self.state.cfr_range = (float(value[0]), float(value[1]))
        else:
            raise ValueError("Unexpected value for 'cfr_range' in settings.")

        value = settings.instance.value(
            "options/loc_precision_range", self.state.loc_precision_range
        )
        if value == "None" or value is None:
            self.state.loc_precision_range = None
        elif type(value) is list:
            self.state.loc_precision_range = (float(value[0]), float(value[1]))
        else:
            raise ValueError("Unexpected value for 'loc_precision_range' in settings.")

        # Read and set 'plot_export_dpi' option
        self.state.plot_export_dpi = int(
            settings.instance.value(
                "options/plot_export_dpi", self.state.plot_export_dpi
            )
        )

        # Read and set 'open_console_at_start' option
        value = settings.instance.value(
            "options/open_console_at_start",
            self.state.open_console_at_start,
        )
        open_console_at_start = (
            value.lower() == "true" if isinstance(value, str) else bool(value)
        )
        self.state.open_console_at_start = open_console_at_start

    def setup_conn(self):
        """Set up signals and slots."""

        # Menu actions
        self.ui.actionLoad.triggered.connect(self.select_and_load_or_import_data_file)
        self.ui.actionSave.triggered.connect(self.save_native_file)
        self.ui.actionExport_data.triggered.connect(self.export_filtered_data)
        self.ui.actionExport_stats.triggered.connect(self.export_filtered_stats)
        self.ui.actionOptions.triggered.connect(self.open_options_dialog)
        self.ui.actionQuit.triggered.connect(self.quit_application)
        self.ui.actionConsole.changed.connect(self.toggle_console_visibility)
        self.ui.actionData_viewer.changed.connect(self.toggle_dataviewer_visibility)
        self.ui.actionState.triggered.connect(self.print_current_state)
        self.ui.actionUnmixer.triggered.connect(self.open_color_unmixer)
        self.ui.actionTime_Inspector.triggered.connect(self.open_time_inspector)
        self.ui.actionAnalyzer.triggered.connect(self.open_analyzer)
        self.ui.actionTrace_Stats_Viewer.triggered.connect(self.open_trace_stats_viewer)
        self.ui.actionTrace_Length_Viewer.triggered.connect(
            self.open_trace_length_viewer
        )
        self.ui.actionFRC_analyzer.triggered.connect(self.open_frc_tool)
        self.ui.actionManual.triggered.connect(
            lambda _: QDesktopServices.openUrl(
                "https://github.com/bsse-scf/pyMINFLUX/wiki/pyMINFLUX-user-manual"
            )
        )
        self.ui.actionWebsite.triggered.connect(
            lambda _: QDesktopServices.openUrl("https://pyminflux.ethz.ch")
        )
        self.ui.actionCode_repository.triggered.connect(
            lambda _: QDesktopServices.openUrl("https://github.com/bsse-scf/pyMINFLUX")
        )
        self.ui.actionIssues.triggered.connect(
            lambda _: QDesktopServices.openUrl(
                "https://github.com/bsse-scf/pyMINFLUX/issues"
            )
        )
        self.ui.actionMailing_list.triggered.connect(
            lambda _: QDesktopServices.openUrl(
                "https://sympa.ethz.ch/sympa/subscribe/pyminflux"
            )
        )
        self.ui.actionWhat_s_new.triggered.connect(
            lambda _: QDesktopServices.openUrl(
                "https://github.com/bsse-scf/pyMINFLUX/blob/master/CHANGELOG.md"
            )
        )
        self.ui.actionCheck_for_updates.triggered.connect(self.check_four_updates)
        self.ui.actionAbout.triggered.connect(self.about)

        # Plotter toolbar
        self.plotter_toolbar.plot_requested_parameters.connect(
            self.plot_selected_parameters
        )
        self.plotter_toolbar.plot_average_positions_state_changed.connect(
            self.full_update_ui
        )

        # Plotter
        self.plotter.locations_selected.connect(
            self.show_selected_points_by_indices_in_dataviewer
        )
        self.plotter.locations_selected_by_range.connect(
            self.show_selected_points_by_range_in_dataviewer
        )
        self.plotter.crop_region_selected.connect(self.crop_data_by_range)
        self.plotter_toolbar.color_code_locs_changed.connect(
            self.plot_selected_parameters
        )

        # Options
        self.options.weigh_avg_localization_by_eco_option_changed.connect(
            self.update_weighted_average_localization_option_and_plot
        )
        self.options.min_trace_length_option_changed.connect(
            self.update_min_trace_length
        )

        # Wizard
        self.wizard.load_data_triggered.connect(
            self.select_and_load_or_import_data_file
        )
        self.wizard.reset_filters_triggered.connect(self.reset_filters_and_broadcast)
        self.wizard.open_unmixer_triggered.connect(self.open_color_unmixer)
        self.wizard.open_time_inspector_triggered.connect(self.open_time_inspector)
        self.wizard.open_analyzer_triggered.connect(self.open_analyzer)
        self.wizard.fluorophore_id_changed.connect(
            self.update_fluorophore_id_in_processor_and_broadcast
        )
        self.wizard.request_fluorophore_ids_reset.connect(self.reset_fluorophore_ids)
        self.wizard.wizard_filters_run.connect(self.full_update_ui)
        self.wizard.save_data_triggered.connect(self.save_native_file)
        self.wizard.export_data_triggered.connect(self.export_filtered_data)
        self.wizard.load_filename_triggered.connect(
            self.select_and_load_or_import_data_file
        )

    def enable_ui_components(self, enabled: bool):
        """Enable/disable UI components."""
        self.plotter.show()  # Always show
        if enabled:
            self.plotter_toolbar.show()
            self.data_viewer.show()
            self.plotter_toolbar.show()
        else:
            self.plotter_toolbar.hide()
            self.data_viewer.hide()
            self.plotter_toolbar.hide()
        self.ui.actionSave.setEnabled(enabled)
        self.ui.actionExport_data.setEnabled(enabled)
        self.ui.actionExport_stats.setEnabled(enabled)
        self.ui.actionUnmixer.setEnabled(enabled)
        self.ui.actionTime_Inspector.setEnabled(enabled)
        self.ui.actionAnalyzer.setEnabled(enabled)
        self.ui.actionTrace_Stats_Viewer.setEnabled(enabled)
        self.ui.actionTrace_Length_Viewer.setEnabled(enabled)
        self.ui.actionFRC_analyzer.setEnabled(enabled)

    def full_update_ui(self):
        """
        Updates the UI completely (after a project load, for instance).
        :return:
        """
        self.plotter.clear()
        self.plot_selected_parameters()
        self.data_viewer.clear()
        if self.processor is not None and self.processor.filtered_dataframe is not None:
            print(f"Retrieved {len(self.processor.filtered_dataframe.index)} events.")

    def print_to_console(self, text):
        """
        Append text to the QTextEdit.
        """
        cursor = self.txConsole.textCursor()
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)
        cursor.insertText(text)
        self.txConsole.setTextCursor(cursor)

    def closeEvent(self, event):
        """Application close event."""

        # Ask the user
        button = QMessageBox.question(
            self,
            f"{__APP_NAME__}",
            "Are you sure you want to quit?",
            QMessageBox.StandardButton.Yes,
            QMessageBox.StandardButton.No,
        )

        if button != QMessageBox.StandardButton.Yes:
            event.ignore()
        else:
            # @TODO Shutdown threads if any are running

            if self.options is not None:
                self.options.close()
                self.options = None

            if self.analyzer is not None:
                self.analyzer.close()
                self.analyzer = None

            if self.color_unmixer is not None:
                self.color_unmixer.close()
                self.color_unmixer = None

            if self.time_inspector is not None:
                self.time_inspector.close()
                self.time_inspector = None

            if self.options is not None:
                self.options.close()
                self.options = None

            if self.trace_stats_viewer is not None:
                self.trace_stats_viewer.close()
                self.trace_stats_viewer = None

            if self.trace_length_viewer is not None:
                self.trace_length_viewer.close()
                self.trace_length_viewer = None

            if self.frc_tool is not None:
                self.frc_tool.close()
                self.frc_tool = None

            # Store the application settings
            if self.last_selected_path != "":
                settings = Settings()
                settings.instance.setValue(
                    "io/last_selected_path", str(self.last_selected_path)
                )

            # # Restore sys.stdout
            # sys.stdout = sys.__stdout__
            # sys.stdout.flush()

            # Now exit
            event.accept()

    @Slot(None, name="reset_filters_and_broadcast")
    def reset_filters_and_broadcast(self):
        """Reset all filters and broadcast changes."""

        # Reset filters and data
        self.processor.reset()
        self.state.efo_thresholds = None
        self.state.cfr_thresholds = None

        # Update main window
        self.full_update_ui()

        # Signal that the external viewers and tools should be updated
        self.request_sync_external_tools.emit()

    @Slot(None, name="quit_application")
    def quit_application(self):
        """Quit the application."""
        self.close()

    @Slot(bool, name="toggle_console_visibility")
    def toggle_console_visibility(self):
        """Toggle the visibility of the console widget."""
        if self.ui.actionConsole.isChecked():
            self.txConsole.show()
        else:
            self.txConsole.hide()

    @Slot(bool, name="toggle_dataviewer_visibility")
    def toggle_dataviewer_visibility(self):
        """Toggle the visibility of the console dock widget."""
        if self.data_viewer is not None:
            if self.ui.actionData_viewer.isChecked():
                self.data_viewer.show()
            else:
                self.data_viewer.hide()

    @Slot(None, name="save_native_file")
    def save_native_file(self):
        """Save data to native pyMINFLUX `.pmx` file."""
        if self.processor is None or len(self.processor.filtered_dataframe.index) == 0:
            return

        # Get current filename to build the suggestion output
        if self.processor.filename is None:
            out_filename = str(self.last_selected_path)
        else:
            out_filename = str(
                self.processor.filename.parent / f"{self.processor.filename.stem}.pmx"
            )

        # Ask the user to pick a name (and format)
        filename, ext = QFileDialog.getSaveFileName(
            self,
            "Save pyMINFLUX dataset",
            out_filename,
            "pyMINFLUX files (*.pmx)",
        )

        # Did the user cancel?
        if filename == "":
            return

        # Does the file name have the .pmx extension?
        if not filename.lower().endswith(".pmx"):
            filename = Path(filename)
            filename = filename.parent / f"{filename.stem}.pmx"

        # Write to disk
        writer = PyMinFluxNativeWriter(self.processor)
        result = writer.write(filename)

        # Save
        if result:
            print(f"Successfully saved {filename}.")
        else:
            QMessageBox.critical(
                self,
                "Error",
                f"Could not save file {Path(filename).name}!\n\nThe error was:\n{writer.message}",
            )

    @Slot(None, name="select_and_load_or_import_data_file")
    def select_and_load_or_import_data_file(self, filename: str = None):
        """
        Pick a MINFLUX `.pmx` file to load, or an Imspector `.npy' or '.mat' file to import.
        :return: void
        """

        # Do we have a filename?
        if filename is None or not filename:
            # Open a file dialog for the user to pick a .pmx, .npy or .mat file
            res = QFileDialog.getOpenFileName(
                self,
                "Load file",
                str(self.last_selected_path),
                "All Supported Files (*.pmx *.npy *.mat);;"
                "pyMINFLUX file (*.pmx);;"
                "Imspector NumPy files (*.npy);;"
                "Imspector MATLAB mat files (*.mat)",
            )
            filename = res[0]

        # Make sure that we are working with a string
        filename = str(filename)

        if filename != "" and Path(filename).is_file():

            # Pick the right reader
            if len(filename) < 5:
                print(f"Invalid file {filename}: skipping.")
                return
            ext = filename.lower()[-4:]

            # Make sure we have a supported file
            if ext not in [".pmx", ".npy", ".mat"]:
                QMessageBox.critical(
                    self,
                    "Error",
                    f"Unsupported file {filename}.",
                )
                return

            # Reload the default settings
            self.load_and_apply_settings()

            # Inform
            print("Reloaded default settings.")

            # If we have a `.pmx` file, we first scan the metadata and update
            # the State
            if ext == ".pmx":
                metadata = NativeMetadataReader.scan(filename)
                if metadata is None:
                    # Could not read the metadata. Abort loading.
                    QMessageBox.critical(
                        self,
                        "Error",
                        f"Could not load {filename}.",
                    )
                    return

                # Update the State from the read metadata
                self.state.update_from_metadata(metadata)

                # Inform
                print("Applied settings from file.")

            # Now pass the filename to the MinFluxReader
            reader = MinFluxReader(
                filename, z_scaling_factor=self.state.z_scaling_factor
            )

            # Open the file
            self.last_selected_path = Path(filename).parent

            # Show some info
            print(reader)

            # Add initialize the processor with the reader
            self.processor = MinFluxProcessor(reader, self.state.min_trace_length)

            # Make sure to set current value of use_weighted_localizations
            self.processor.use_weighted_localizations = (
                self.state.weigh_avg_localization_by_eco
            )

            # Show the filename on the main window
            self.setWindowTitle(
                f"{__APP_NAME__} v{__version__} - [{Path(filename).name}]"
            )

            # Close the Options
            if self.options is not None:
                self.options.close()
                self.options = None

            # Close the Color Unmixer
            if self.color_unmixer is not None:
                self.color_unmixer.close()
                self.color_unmixer = None

            # Close the Temporal Inspector
            if self.time_inspector is not None:
                self.time_inspector.close()
                self.time_inspector = None

            # Close the Trace Stats Viewer
            if self.trace_stats_viewer is not None:
                self.trace_stats_viewer.close()
                self.trace_stats_viewer = None

            if self.trace_length_viewer is not None:
                self.trace_length_viewer.close()
                self.trace_length_viewer = None

            # Close the FRC Tool
            if self.frc_tool is not None:
                self.frc_tool.close()
                self.frc_tool = None

            # Update the ui
            self.full_update_ui()

            # Make sure to autoupdate the axis on load
            self.plotter.enableAutoRange(enable=True)

            # Reset the fluorophore list in the wizard
            self.wizard.set_fluorophore_list(self.processor.num_fluorophores)

            # Enable selected ui components
            self.enable_ui_components(True)

            # Update the Analyzer
            if self.analyzer is not None:
                self.analyzer.set_processor(self.processor)
                self.analyzer.plot()

            # Attach the processor reference to the wizard
            self.wizard.set_processor(self.processor)

            # Enable wizard
            self.wizard.enable_controls(True)

        else:
            # If nothing is loaded (even from earlier), disable wizard
            if self.processor is None:
                self.wizard.enable_controls(False)

    @Slot(None, name="export_filtered_data")
    def export_filtered_data(self):
        """Export filtered data as CSV file."""
        if self.processor is None or len(self.processor.filtered_dataframe.index) == 0:
            return

        # Get current filename to build the suggestion output
        if self.processor.filename is None:
            out_filename = str(self.last_selected_path)
        else:
            out_filename = str(
                self.processor.filename.parent / f"{self.processor.filename.stem}.csv"
            )

        # Ask the user to pick a name (and format)
        filename, ext = QFileDialog.getSaveFileName(
            self,
            "Export filtered data",
            out_filename,
            "Comma-separated value files (*.csv)",
        )

        # Did the user cancel?
        if filename == "":
            return

        if ext == "Comma-separated value files (*.csv)":
            # Does the file name have the .csv extension?
            if not filename.lower().endswith(".csv"):
                filename = Path(filename)
                filename = filename.parent / f"{filename.stem}.csv"

            # Write to disk
            result = MinFluxWriter.write_csv(self.processor, filename)

        else:
            return

        # Save
        if result:
            print(
                f"Successfully exported {len(self.processor.filtered_dataframe.index)} localizations."
            )
        else:
            QMessageBox.critical(
                self,
                "Error",
                f"Could not export filtered data to {Path(filename).name}.",
            )

    @Slot(None, name="export_filtered_stats")
    def export_filtered_stats(self):
        """Export filtered, per-trace statistics as CSV file."""
        if self.processor is None or len(self.processor.filtered_dataframe.index) == 0:
            # Inform and return
            QMessageBox.information(
                self,
                "Info",
                f"Sorry, nothing to export.",
            )
            return

        # Get current filename to build the suggestion output
        if self.processor.filename is None:
            out_filename = str(self.last_selected_path)
        else:
            out_filename = str(
                self.processor.filename.parent
                / f"{self.processor.filename.stem}_stats.csv"
            )

        # Ask the user to pick a name
        filename, ext = QFileDialog.getSaveFileName(
            self,
            "Export trace statistics",
            out_filename,
            "Comma-separated value files (*.csv)",
        )

        # Did the user cancel?
        if filename == "":
            return

        # Does the file name have the .csv extension?
        if ext != "Comma-separated value files (*.csv)":
            return

        if not filename.lower().endswith(".csv"):
            filename = Path(filename)
            filename = filename.parent / f"{filename.stem}.csv"

        # Collect stats
        stats = self.processor.filtered_dataframe_stats
        if stats is None:
            return

        # Write stats to disk
        try:
            stats.to_csv(filename, index=False)
        except Exception as e:
            QMessageBox.critical(
                self,
                "Error",
                f"Could not export trace statistics to {Path(filename).name}.\n\n{str(e)}.",
            )
            return

        print(f"Successfully exported statistics for {len(stats.index)} traces.")

    @Slot(None, name="print_current_state")
    def print_current_state(self):
        """Print current contents of the state machine (DEBUG)."""
        if self.txConsole.isHidden():
            self.txConsole.show()
            self.ui.actionConsole.setChecked(True)
        state_dict = self.state.asdict()
        print("[DEBUG] Internal state:")
        for s in state_dict:
            print(f'  ∟ "{s}": {state_dict[s]}')

    @Slot(None, name="about")
    def about(self):
        """Show simple About dialog."""
        QMessageBox.about(
            self,
            "About",
            f"{__APP_NAME__} v{__version__}\n"
            f"\n"
            f"Copyright 2022 - {datetime.now().year}\n"
            f"Single Cell Facility\n"
            f"D-BSSE\n"
            f"ETH Zurich\n"
            f"Switzerland",
        )

    @Slot(None, name="check_four_updates")
    def check_four_updates(self):
        """Check for application updates."""

        # Check for updates
        code, version, error = check_for_updates()

        # Process the output
        if code == -1:
            # Something went wrong: report
            html = f"<b>Error! {error}</b><br /><br /><br />Please make sure you are connected to the internet.<br />If this error persists, please <a href='https://github.com/bsse-scf/pyMINFLUX/issues/'>report it</a>."

        elif code == 0:
            # No new version
            html = f"<b>Congratulations!</b><br /><br />You are running the latest version ({pyminflux.__version__}) of {pyminflux.__APP_NAME__}."

        elif code == 1:
            # Show a dialog with a link to the download page
            html = f"<b>There is a new version ({version}) of {pyminflux.__APP_NAME__}!</b><br /><br />You can download it from the <a href='https://github.com/bsse-scf/pyMINFLUX/releases/latest'>release page</a>."

        else:
            raise ValueError("Unexpected code!")

        # Show the dialog
        dialog = QDialog()
        dialog.setWindowTitle("Check for updates")
        dialog.setMinimumSize(400, 180)
        dialog.setFixedHeight(180)
        layout = QVBoxLayout(dialog)
        text_browser = QTextBrowser()
        text_browser.setStyleSheet("background-color: transparent;")
        text_browser.setOpenExternalLinks(True)
        text_browser.insertHtml(html)
        layout.addWidget(text_browser)
        button = QPushButton("OK")
        button.clicked.connect(dialog.close)
        layout.addWidget(button)
        dialog.exec_()

    @Slot(None, name="open_analyzer")
    def open_analyzer(self):
        """Initialize and open the analyzer."""
        if self.analyzer is None:
            self.analyzer = Analyzer(self.processor)
            self.wizard.wizard_filters_run.connect(self.analyzer.plot)
            self.request_sync_external_tools.connect(self.analyzer.plot)
            self.wizard.efo_bounds_modified.connect(self.analyzer.change_efo_bounds)
            self.wizard.cfr_bounds_modified.connect(self.analyzer.change_cfr_bounds)
            self.analyzer.data_filters_changed.connect(self.full_update_ui)
            self.analyzer.cfr_threshold_factor_changed.connect(
                self.wizard.change_cfr_threshold_factor
            )
            self.analyzer.efo_bounds_changed.connect(self.wizard.change_efo_bounds)
            self.analyzer.cfr_bounds_changed.connect(self.wizard.change_cfr_bounds)
            if self.time_inspector is not None:
                self.analyzer.data_filters_changed.connect(self.time_inspector.update)
            if self.trace_stats_viewer is not None:
                self.analyzer.data_filters_changed.connect(
                    self.trace_stats_viewer.update
                )
            if self.trace_length_viewer is not None:
                self.analyzer.data_filters_changed.connect(
                    self.trace_length_viewer.update
                )
            self.analyzer.plot()
        self.analyzer.show()
        self.analyzer.activateWindow()

    @Slot(None, name="open_time_inspector")
    def open_time_inspector(self):
        """Initialize and open the Time Inspector."""
        if self.time_inspector is None:
            self.time_inspector = TimeInspector(self.processor, parent=self)
            self.time_inspector.dataset_time_filtered.connect(self.full_update_ui)
            self.wizard.wizard_filters_run.connect(self.time_inspector.update)
            self.request_sync_external_tools.connect(self.time_inspector.update)
        if self.analyzer is not None:
            self.analyzer.data_filters_changed.connect(self.time_inspector.update)
            self.time_inspector.dataset_time_filtered.connect(self.analyzer.plot)
        if self.trace_stats_viewer is not None:
            self.time_inspector.dataset_time_filtered.connect(
                self.trace_stats_viewer.update
            )
        if self.trace_length_viewer is not None:
            self.time_inspector.dataset_time_filtered.connect(
                self.trace_length_viewer.update
            )
        self.time_inspector.show()
        self.time_inspector.activateWindow()

    @Slot(None, name="open_color_unmixer")
    def open_color_unmixer(self):
        """Initialize and open the color unmixer."""
        if self.color_unmixer is None:
            self.color_unmixer = ColorUnmixer(self.processor, parent=self)
            self.color_unmixer.fluorophore_ids_assigned.connect(
                self.wizard.set_fluorophore_list
            )
            self.color_unmixer.fluorophore_ids_assigned.connect(
                self.plot_selected_parameters
            )
            self.wizard.wizard_filters_run.connect(self.plot_selected_parameters)
        if self.trace_stats_viewer is not None:
            self.color_unmixer.fluorophore_ids_assigned.connect(
                self.trace_stats_viewer.update
            )
        if self.trace_length_viewer is not None:
            self.color_unmixer.fluorophore_ids_assigned.connect(
                self.trace_length_viewer.update
            )
        self.color_unmixer.show()
        self.color_unmixer.activateWindow()

    @Slot(None, name="open_options_dialog")
    def open_options_dialog(self):
        """Open the options dialog."""
        if self.options is None:
            self.options = Options()
            self.options.min_trace_length_option_changed.connect(
                self.update_min_trace_length
            )
        self.options.show()
        self.options.activateWindow()

    @Slot(None, name="update_min_trace_length")
    def update_min_trace_length(self):
        if self.processor is not None:
            self.processor.min_trace_length = self.state.min_trace_length

    @Slot(None, name="open_trace_stats_viewer")
    def open_trace_stats_viewer(self):
        """Open the trace stats viewer."""
        if self.trace_stats_viewer is None:
            self.trace_stats_viewer = TraceStatsViewer(self.processor, parent=self)
            self.request_sync_external_tools.connect(self.trace_stats_viewer.update)
            self.wizard.request_fluorophore_ids_reset.connect(
                self.trace_stats_viewer.update
            )
            self.wizard.wizard_filters_run.connect(self.trace_stats_viewer.update)
            self.trace_stats_viewer.export_trace_stats_requested.connect(
                self.export_filtered_stats
            )
            self.wizard.fluorophore_id_changed.connect(self.trace_stats_viewer.update)
        if self.color_unmixer is not None:
            self.color_unmixer.fluorophore_ids_assigned.connect(
                self.trace_stats_viewer.update
            )
        if self.analyzer is not None:
            self.analyzer.data_filters_changed.connect(self.trace_stats_viewer.update)
        self.trace_stats_viewer.show()
        self.trace_stats_viewer.activateWindow()

    @Slot(None, name="open_trace_length_viewer")
    def open_trace_length_viewer(self):
        """Open the trace length viewer."""
        if self.trace_length_viewer is None:
            self.trace_length_viewer = TraceLengthViewer(self.processor, parent=self)
            self.request_sync_external_tools.connect(self.trace_length_viewer.update)
            self.wizard.request_fluorophore_ids_reset.connect(
                self.trace_length_viewer.update
            )
            self.wizard.wizard_filters_run.connect(self.trace_length_viewer.update)
            self.wizard.fluorophore_id_changed.connect(self.trace_length_viewer.update)
        if self.color_unmixer is not None:
            self.color_unmixer.fluorophore_ids_assigned.connect(
                self.trace_length_viewer.update
            )
        if self.analyzer is not None:
            self.analyzer.data_filters_changed.connect(self.trace_length_viewer.update)
        self.trace_length_viewer.show()
        self.trace_length_viewer.activateWindow()

    @Slot(None, name="open_frc_tool")
    def open_frc_tool(self):
        """Open the FRC tool."""
        if self.frc_tool is None:
            self.frc_tool = FRCTool(self.processor)
        self.frc_tool.show()
        self.frc_tool.activateWindow()

    @Slot(list, name="show_selected_points_by_indices_in_dataviewer")
    def show_selected_points_by_indices_in_dataviewer(self, points):
        """Show the data for the selected points in the dataframe viewer."""

        # Extract indices of the rows corresponding to the selected points
        indices = []
        for p in points:
            indices.append(p.index())

        # Sort the indices
        indices = sorted(indices)

        # Get the filtered dataframe subset corresponding to selected indices
        df = self.processor.select_by_indices(
            indices=indices, from_weighted_locs=self.state.plot_average_localisations
        )

        # Update the dataviewer
        self.data_viewer.set_data(df)

        # Inform
        point_str = "event" if len(indices) == 1 else "events"
        print(f"Selected {len(indices)} {point_str}.")

    @Slot(tuple, tuple, name="show_selected_points_by_range_in_dataviewer")
    def show_selected_points_by_range_in_dataviewer(
        self, x_param, y_param, x_range, y_range
    ):
        """Select the data by x and y range and show in the dataframe viewer."""

        # Get the filtered dataframe subset contained in the provided x and y ranges
        df = self.processor.select_by_2d_range(
            x_param,
            y_param,
            x_range,
            y_range,
            from_weighted_locs=self.state.plot_average_localisations,
        )

        # Update the dataviewer
        self.data_viewer.set_data(df)

        # Inform
        point_str = "event" if len(df.index) == 1 else "events"
        print(f"Selected {len(df.index)} {point_str}.")

    @Slot(tuple, tuple, name="crop_data_by_range")
    def crop_data_by_range(self, x_param, y_param, x_range, y_range):
        """Filter the data by x and y range and show in the dataframe viewer."""

        # Filter the dataframe by the passed x and y ranges
        self.processor.filter_by_2d_range(x_param, y_param, x_range, y_range)

        # Update the Analyzer
        if self.analyzer is not None:
            self.analyzer.plot()

        # Update the Fluorophore Detector?
        if self.color_unmixer is not None:
            # No need to update
            pass

        # Update the Temporal Inspector?
        if self.time_inspector is not None:
            # No need to update
            pass

        # Update the ui
        self.full_update_ui()

        # Make sure to autoupdate the axis (on load only)
        self.plotter.getViewBox().enableAutoRange(axis=ViewBox.XYAxes, enable=True)

        # Signal that the external viewers and tools should be updated
        self.request_sync_external_tools.emit()

    def update_weighted_average_localization_option_and_plot(self):
        """Update the weighted average localization option in the Processor and re-plot."""
        if self.processor is not None:
            self.processor.use_weighted_localizations = (
                self.state.weigh_avg_localization_by_eco
            )
        self.plot_selected_parameters()

    def plot_selected_parameters(self):
        """Plot the localizations."""

        # Remove the previous plots
        self.plotter.remove_points()

        # If there is nothing to plot, return here
        if self.processor is None:
            return

        # If an only if the requested parameters are "x" and "y" (in any order),
        # we consider the State.plot_average_localisations property.
        if (self.state.x_param == "x" and self.state.y_param == "y") or (
            self.state.x_param == "y" and self.state.y_param == "x"
        ):
            if self.state.plot_average_localisations:
                # Get the (potentially filtered) averaged dataframe
                dataframe = self.processor.weighted_localizations
            else:
                # Get the (potentially filtered) full dataframe
                dataframe = self.processor.filtered_dataframe

        else:
            # Get the (potentially filtered) full dataframe
            dataframe = self.processor.filtered_dataframe

        # Extract values
        x = dataframe[self.state.x_param].values
        y = dataframe[self.state.y_param].values
        tid = dataframe["tid"].values
        fid = dataframe["fluo"].values

        # Always plot the (x, y) coordinates in the 2D plotter
        self.plotter.plot_parameters(
            tid=tid,
            fid=fid,
            x=x,
            y=y,
            x_param=self.state.x_param,
            y_param=self.state.y_param,
        )

    def show_processed_dataframe(self, dataframe=None):
        """
        Displays the results for current frame in the data viewer.
        """

        # Is there data to process?
        if self.processor is None:
            self.data_viewer.clear()
            return

        if dataframe is None:
            # Get the (potentially filtered) dataframe
            dataframe = self.processor.filtered_dataframe()

        # Pass the dataframe to the pdDataViewer
        self.data_viewer.set_data(dataframe)

    @Slot(int, name="update_fluorophore_id_in_processor_and_broadcast")
    def update_fluorophore_id_in_processor_and_broadcast(self, index):
        """Update the fluorophore ID in the processor and broadcast the change to all parties."""

        # Update the processor
        self.processor.current_fluorophore_id = index

        # Update all views
        self.full_update_ui()

        # Update the analyzer as well
        if self.analyzer is not None:
            self.analyzer.plot()

    def reset_fluorophore_ids(self):
        """Reset the fluorophore IDs."""

        # Reset
        self.processor.reset()

        # Update UI
        self.update_fluorophore_id_in_processor_and_broadcast(0)



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import sys

from PySide6.QtCore import QObject, Signal


class EmittingStream(QObject):
    """
    Redirect standard output and error to be displayed by a QWidget.
    """

    signal_textWritten = Signal(str, name="signal_textWritten")

    def write(self, text):
        self.signal_textWritten.emit(str(text))

    def flush(self):
        sys.stdout = sys.__stdout__
        sys.stdout.flush()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import pandas as pd
from PySide6 import QtCore
from PySide6.QtCore import Qt


class PandasDataModel(QtCore.QAbstractTableModel):
    """
    Class to populate a table view with a pandas dataframe
    """

    def __init__(self, data, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)
        self._data = data

    def clear(self):
        # Begin a batch operation to remove all rows
        self.beginResetModel()

        # Clear your data here
        self._data = pd.DataFrame()

        # End the batch operation, notifying the views
        self.endResetModel()

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parent=None):
        return self._data.shape[1]

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                value = self._data.iloc[index.row(), index.column()]
                if self._data.columns[index.column()] in ["efo", "dwell"]:
                    return f"{int(value)}"
                if self._data.columns[index.column()] in ["x", "y", "z"]:
                    return f"{value:.1f}"
                if self._data.columns[index.column()] in ["cfr", "dcr"]:
                    return f"{value:.3f}"
                if isinstance(value, float):  # Check if the value is a float
                    return (
                        f"{value:.2f}"  # Format the float value with 2 decimal places
                    )
                return str(value)
        return None

    def headerData(self, col, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._data.columns[col]
        return None

    def sort(self, column, order):
        self.layoutAboutToBeChanged.emit()
        column_name = self._data.columns[column]
        self._data.sort_values(
            column_name, ascending=(order == Qt.AscendingOrder), inplace=True
        )
        self.layoutChanged.emit()
